from datetime import date, datetime

from django.db import models
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers as rest_serializers

from wbcore import serializers
from wbcore.models import WBModel


class Data(WBModel):
    """A blueprint for what data can be attached to a workflow."""

    class DataType(models.TextChoices):
        CHAR = "Text", _("Text")
        INT = "Number", _("Number")
        DATE = "Date", _("Date")
        DATETIME = "Date Time", _("Date Time")
        BOOL = "Boolean", _("Boolean")

        @classmethod
        def get_serializer_field_mapping(cls) -> dict:
            serializer_fields = [
                serializers.CharField(),
                serializers.IntegerField(),
                serializers.DateField(),
                serializers.DateTimeField(),
                serializers.BooleanField(),
            ]
            return {
                data_type: serializer_field
                for data_type, serializer_field in zip(cls, serializer_fields, strict=False)
            }

        @classmethod
        def get_cast_mapping(cls) -> dict:
            cast_callables = [
                str,
                int,
                datetime.strptime,
                datetime.strptime,
                bool,
            ]
            return {data_type: cast_callable for data_type, cast_callable in zip(cls, cast_callables, strict=False)}

    workflow = models.ForeignKey(
        to="workflow.Workflow",
        on_delete=models.CASCADE,
        verbose_name=_("Workflow"),
        related_name="attached_data",
    )
    label = models.CharField(max_length=64, verbose_name=_("Label"))
    help_text = models.CharField(max_length=128, verbose_name=_("Help Text"), null=True, blank=True)
    data_type = models.CharField(
        max_length=64, choices=DataType.choices, verbose_name=_("Data Type"), default=DataType.CHAR
    )
    required = models.BooleanField(
        verbose_name=_("Required"),
    )
    default = models.CharField(
        max_length=128,
        verbose_name=_("Default"),
        blank=True,
        null=True,
    )

    def __str__(self) -> str:
        return f"{self.label} ({self.workflow})"

    def get_serializer_field(self) -> rest_serializers.Field:
        """Constructs a serializer field from the instance

        Returns:
            rest_serializers.Field: A serializer field based on the instance

        Raises:
            ValueError: If casting the default value fails
        """

        field = self.DataType.get_serializer_field_mapping()[self.data_type]
        field.set_attr("label", self.label)
        field.set_attr("required", self.required)
        if self.default:
            casted_value = Data.cast_value_to_datatype(self.data_type, self.default)
            field.set_attr("default", casted_value)
        if self.help_text:
            field.set_attr("help_text", self.help_text)
        return field

    @classmethod
    def cast_value_to_datatype(cls, data_type: DataType, value: str) -> datetime | str | int | bool:
        """Casts a value string into a provided data type

        Args:
            data_type (DataType): Selected data type
            value (str): The value string

        Returns:
            datetime | str | int | bool: Value string casted into selected data type

        Raises:
            ValueError: If casting fails
        """

        cast_callable = cls.DataType.get_cast_mapping()[data_type]
        if data_type == cls.DataType.DATE:
            # NOTE: Different format? ISO-8601? Create global dateime format preference?
            casted_value: date = cast_callable.__call__(value, "%d.%m.%Y").date()
        elif data_type == cls.DataType.DATETIME:
            casted_value: datetime = cast_callable.__call__(value, "%d.%m.%Y %H:%M:%S")
        elif data_type == cls.DataType.BOOL:
            if value.lower() == "true":
                casted_value: bool = True
            elif value.lower() == "false":
                casted_value: bool = False
            else:
                raise ValueError(gettext("Casting to bool failed."))
        else:
            casted_value: str | int = cast_callable.__call__(value)
        return casted_value

    @classmethod
    def cast_value_from_target_object(
        cls, target_type_object: datetime | date | bool | str | int, input_str: str
    ) -> datetime | date | bool | str | int | None:
        """Casts an input string into the type of a provided object. Supported data types: datetime | date | bool | str | int

        Args:
            target_type_object (datetime | date | bool | str | int): The target type object
            input_str (str): An input string to be casted

        Raises:
            ValueError: When an unsupported type is provided or casting failed

        Returns:
            datetime | date | bool | str | int | None: The casted input string
        """

        casted_output = None

        if isinstance(target_type_object, datetime):
            data_type = cls.DataType.DATETIME
        elif isinstance(target_type_object, date):
            data_type = cls.DataType.DATE
        elif isinstance(target_type_object, bool):
            data_type = cls.DataType.BOOL
        elif isinstance(target_type_object, int):
            data_type = cls.DataType.INT
        elif isinstance(target_type_object, str):
            data_type = cls.DataType.CHAR
        else:
            raise ValueError(gettext("Unsupported target type!"))

        casted_output = cls.cast_value_to_datatype(data_type, input_str)

        return casted_output

    @classmethod
    def cast_datatype_to_string(cls, data_type: DataType, data_object: datetime | str | int | bool) -> str:
        """Casts an object of type datatype into a str. Reverse function of 'cast_value_to_datatype'

        Args:
            data_type (DataType): Selected data type
            data_object (datetime | str | int | bool): The object to be casted

        Returns:
            str: A string representation of the object

        Raises:
            ValueError: If casting fails
        """

        if data_type in [cls.DataType.DATE, cls.DataType.DATETIME]:
            if data_type == cls.DataType.DATE:
                # NOTE: Different format? ISO-8601? Create global dateime format preference?
                format_str = "%d.%m.%Y"
            else:
                format_str = "%d.%m.%Y %H:%M:%S"
            try:
                casted_object = data_object.strftime(format_str)
            except AttributeError as e:
                raise ValueError(gettext("Date(time) type selected but no date(time) object provided!")) from e

        elif data_type == cls.DataType.BOOL:
            if data_object is True:
                casted_object = "True"
            elif data_object is False:
                casted_object = "False"
            else:
                raise ValueError(gettext("Wrong boolean value!"))
        else:
            casted_object = str(data_object)
        return casted_object

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:data"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:datarepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{label}}"

    class Meta:
        verbose_name = _("Data")
        verbose_name_plural = _("Data")
        constraints = [
            models.UniqueConstraint(fields=["label", "workflow"], name="unique_label_workflow"),
        ]


class DataValue(models.Model):
    """A model that holds the value for a data object"""

    value = models.CharField(max_length=64, verbose_name=_("Value"), blank=True, null=True)
    data = models.ForeignKey(to=Data, related_name="values", on_delete=models.CASCADE, verbose_name=_("Data"))
    process = models.ForeignKey(
        to="workflow.Process", related_name="data_values", on_delete=models.CASCADE, verbose_name=_("Process")
    )

    def __str__(self) -> str:
        return f"{self.data}: {self.value}"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{value}}"

    class Meta:
        verbose_name = _("Data Value")
        verbose_name_plural = _("Data Values")



import phonenumbers
from django.core.validators import validate_email
from django.forms import ValidationError
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from dynamic_preferences.registries import global_preferences_registry
from rest_framework.reverse import reverse
from slugify import slugify

from wbcore import serializers as wb_serializers
from wbcore.contrib.geography.serializers import GeographyRepresentationSerializer

from ..models import (
    ClientManagerRelationship,
    Company,
    CompanyType,
    CustomerStatus,
    EmailContact,
    Entry,
    Person,
    Specialization,
    TelephoneContact,
)
from .contacts import SocialMediaContactRepresentationSerializer
from .entry_representations import EntryRepresentationSerializer


class CompanyRepresentationSerializer(EntryRepresentationSerializer):
    _detail_preview = wb_serializers.HyperlinkField(reverse_name="wbcore:directory:company-detail")
    profile_image = wb_serializers.ImageField(required=False, read_only=True)

    class Meta:
        model = Company
        fields = (
            "id",
            "computed_str",
            "_detail",
            "_detail_preview",
            "primary_email",
            "primary_telephone",
            "profile_image",
            "_additional_resources",
        )


class PersonRepresentationSerializer(EntryRepresentationSerializer):
    _detail_preview = wb_serializers.HyperlinkField(reverse_name="wbcore:directory:person-detail")
    profile_image = wb_serializers.ImageField(required=False, read_only=True)

    class Meta:
        model = Person
        fields = (
            "id",
            "computed_str",
            "first_name",
            "last_name",
            "_detail",
            "_detail_preview",
            "primary_email",
            "primary_telephone",
            "profile_image",
            "_additional_resources",
        )


class InternalUserProfileRepresentationSerializer(PersonRepresentationSerializer):
    def get_filter_params(self, request):
        return {"only_internal_users": True}


class CompanyTypeModelSerializer(wb_serializers.ModelSerializer):
    class Meta:
        model = CompanyType
        fields = (
            "id",
            "title",
        )

    def validate(self, data):
        title = data.get("title", None)
        if title:
            type = CompanyType.objects.filter(slugify_title=slugify(title, separator=" "))
            if obj := self.instance:
                type = type.exclude(id=obj.id)
            if type.exists():
                raise ValidationError({"title": _("Cannot add a duplicate company type.")})
        return data


class SpecializationModelSerializer(wb_serializers.ModelSerializer):
    def validate(self, data):
        title = data.get("title", None)
        if title:
            specialization = Specialization.objects.filter(slugify_title=slugify(title, separator=" "))
            if obj := self.instance:
                specialization = specialization.exclude(id=obj.id)
            if specialization.exists():
                raise ValidationError({"title": _("Cannot add a duplicate specialization.")})
        return data

    class Meta:
        model = Specialization
        fields = (
            "id",
            "title",
        )


class CustomerStatusModelSerializer(wb_serializers.ModelSerializer):
    class Meta:
        model = CustomerStatus
        fields = (
            "id",
            "title",
        )

    def validate(self, data):
        title = data.get("title", None)
        if title:
            status = CustomerStatus.objects.filter(slugify_title=slugify(title, separator=" "))
            if obj := self.instance:
                status = status.exclude(id=obj.id)
            if status.exists():
                raise ValidationError({"title": _("Cannot add a duplicate customer status.")})
        return data


class CompanyTypeRepresentationSerializer(wb_serializers.RepresentationSerializer):
    endpoint = "wbcore:directory:companytyperepresentation-list"
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:directory:companytype-detail")

    class Meta:
        model = CompanyType
        fields = (
            "id",
            "title",
            "_detail",
        )


class CustomerStatusRepresentationSerializer(wb_serializers.RepresentationSerializer):
    endpoint = "wbcore:directory:customerstatusrepresentation-list"
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:directory:customerstatus-detail")

    class Meta:
        model = CustomerStatus
        fields = (
            "id",
            "title",
            "_detail",
        )


class SpecializationRepresentationSerializer(wb_serializers.RepresentationSerializer):
    endpoint = "wbcore:directory:specializationrepresentation-list"
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:directory:specialization-detail")

    class Meta:
        model = Specialization
        fields = (
            "id",
            "title",
            "_detail",
        )


class EntryModelSerializer(wb_serializers.ModelSerializer):
    last_event = wb_serializers.CharField(read_only=True)
    last_event_period_endswith = wb_serializers.DateTimeField(read_only=True)
    activity_heat = wb_serializers.RangeSelectField(
        color="rgb(220,20,60)", label=_("Activity Heat"), read_only=True, required=False
    )
    cities = wb_serializers.PrimaryKeyRelatedField(read_only=True)
    _cities = GeographyRepresentationSerializer(source="cities", many=True, read_only=True)

    customer_status = wb_serializers.CharField(
        help_text=_(
            "This field indicates a person's primary employer's status. It is automatically filled when the status of the primary employer is set."
        ),
        label=_("Customer Status"),
        read_only=True,
        required=False,
    )
    is_primary_employer = wb_serializers.BooleanField(read_only=True)
    position_in_company = wb_serializers.CharField(
        allow_null=True,
        default="",
        help_text=_(
            "This field indicates a person's position with their primary employer. It is automatically filled when this person is assigned an employer and this person's position with that employer is filled in."
        ),
        label=_("Position In Company"),
        read_only=True,
        required=False,
    )
    primary_address = wb_serializers.CharField(
        allow_null=True, read_only=True, required=False, label=_("Primary Address")
    )
    primary_email = wb_serializers.CharField(
        allow_null=True, required=False, read_only=False, label=_("Primary Email")
    )
    primary_manager_repr = wb_serializers.CharField(
        allow_null=True, read_only=True, required=False, label=_("Primary Manager")
    )
    primary_telephone = wb_serializers.TelephoneField(
        allow_null=True,
        required=False,
        label=_("Primary Telephone"),
    )
    primary_website = wb_serializers.URLField(allow_null=True, required=False, label=_("Primary Website"))
    primary_social = wb_serializers.URLField(allow_null=True, required=False, label=_("Primary Social"))
    profile_image = wb_serializers.ImageField(allow_null=True, required=False, label=_("Profile Image"))
    _relationship_managers = PersonRepresentationSerializer(many=True, source="relationship_managers")
    _social_media = SocialMediaContactRepresentationSerializer(many=True, read_only=True, source="social_media")
    primary_manager = wb_serializers.PrimaryKeyRelatedField(label=_("Primary Manager"), required=False, read_only=True)
    _primary_manager = InternalUserProfileRepresentationSerializer(source="primary_manager")

    @wb_serializers.register_resource()
    def detail(self, instance: Entry, request, user):
        if not (request := self.context.get("request")):
            return
        if instance.is_company:
            reverse_url = reverse("wbcore:directory:company-detail", args=[instance.id], request=request)
        else:
            reverse_url = reverse("wbcore:directory:person-detail", args=[instance.id], request=request)
        return {"detail": reverse_url}

    @wb_serializers.register_resource()
    def relationships(self, instance, request, user):
        if user.has_perm("directory.view_relationship"):
            return {
                "relationships": reverse(
                    "wbcore:directory:entry-relationship-list", args=[instance.id], request=request
                )
            }
        return {}

    @wb_serializers.register_resource()
    def get_managers(self, instance, request, user):
        return {
            "manager": f"{reverse('wbcore:directory:clientmanagerrelationship-list', args=[], request=request)}?client={instance.id}"
        }

    @wb_serializers.register_resource()
    def contacts(self, instance, request, user):
        crms = {
            "addresses": reverse("wbcore:directory:entry-addresscontact-list", args=[instance.id], request=request),
            "bankings": reverse("wbcore:directory:entry-bankingcontact-list", args=[instance.id], request=request),
            "emails": reverse("wbcore:directory:entry-emailcontact-list", args=[instance.id], request=request),
            "telephones": reverse("wbcore:directory:entry-telephonecontact-list", args=[instance.id], request=request),
            "social_media": reverse(
                "wbcore:directory:entry-socialmediacontact-list", args=[instance.id], request=request
            ),
            "websites": reverse("wbcore:directory:entry-websitecontact-list", args=[instance.id], request=request),
        }

        return crms

    def validate(self, data):
        if primary_email := data.get("primary_email", None):
            try:
                validate_email(primary_email)
            except ValidationError as e:
                raise ValidationError({"primary_email": "Invalid e-mail address"}) from e

        if primary_telephone := data.get("primary_telephone", None):
            try:
                parser_number = phonenumbers.parse(
                    primary_telephone, global_preferences_registry.manager()["directory__main_country_code"]
                )
                if parser_number:
                    formatted_number = phonenumbers.format_number(parser_number, phonenumbers.PhoneNumberFormat.E164)
                    data["primary_telephone"] = formatted_number
                else:
                    raise ValidationError({"primary_telephone": gettext("Invalid phone number format")})
            except Exception as e:
                raise ValidationError({"primary_telephone": gettext("Invalid phone number format")}) from e
        return super().validate(data)

    def update(self, instance, validated_data):
        if "primary_email" in validated_data.keys():
            primary_email = validated_data.get("primary_email", None)
            if primary_email:
                EmailContact.set_entry_primary_email(instance, primary_email)
            else:
                instance.emails.filter(primary=True).delete()

        if "primary_telephone" in validated_data.keys():
            primary_telephone = validated_data.get("primary_telephone", None)
            if primary_telephone:
                TelephoneContact.set_entry_primary_telephone(instance, primary_telephone)
            else:
                instance.telephones.filter(primary=True).delete()
        return super().update(instance, validated_data)

    def create(self, validated_data: dict) -> Entry:
        primary_email = validated_data.pop("primary_email", None)
        primary_telephone = validated_data.pop("primary_telephone", None)
        primary_manager = validated_data.pop("primary_manager", None)
        instance = super().create(validated_data)
        if primary_email:
            EmailContact.set_entry_primary_email(instance, primary_email)
        if primary_telephone:
            TelephoneContact.set_entry_primary_telephone(instance, primary_telephone)

        if primary_manager:
            manager: Person = primary_manager
        elif request := self.context.get("request"):
            manager: Person = request.user.profile
        else:
            manager = None

        if manager:
            ClientManagerRelationship.objects.create(
                client=instance,
                relationship_manager=manager,
                status=ClientManagerRelationship.Status.PENDINGADD,
                primary=True,
            )
        return instance

    class Meta:
        model = Entry
        read_only_fields = ("entry_type",)

        fields = (
            "last_event",
            "last_event_period_endswith",
            "id",
            "computed_str",
            "activity_heat",
            "addresses",
            "cities",
            "_cities",
            "customer_status",
            "entry_type",
            "is_draft_entry",
            "is_primary_employer",
            "position_in_company",
            "primary_address",
            "primary_email",
            "primary_manager_repr",
            "primary_telephone",
            "primary_website",
            "primary_social",
            "profile_image",
            "relationship_managers",
            "_relationship_managers",
            "salutation",
            "social_media",
            "_social_media",
            "signature",
            "_additional_resources",
            "primary_manager",
            "_primary_manager",
        )



class DeserializationError(Exception):
    """
    Exception excepted when a deserialized error happens during the process object phase in the Handler.

    This exception won't stop the importing process and will only be logged
    """

    pass


class SkipImportError(Exception):
    """
    Exception excepted when a deserialized skip happens during the process object phase in the Handler.

    This exception won't stop the importing process and won't trigger a warning
    """


class ImportError(Exception):
    """
    Exception returns when something wrong happens during the processing of the import source and means that not all data were succesfully imported
    """

    pass



import pytest
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from guardian.shortcuts import assign_perm
from wbcore.contrib.authentication.models.users import Permission
from wbcore.contrib.guardian.viewsets import PivotUserObjectPermissionModelViewSet


class TestPivotUserObjectPermissionModelViewSet:
    def test_cached_property_permissions(self, mocker, request):
        mocked_filter = mocker.patch("wbcore.contrib.guardian.viewsets.viewsets.Permission.objects.filter")
        view = PivotUserObjectPermissionModelViewSet(request=request, kwargs={"content_type_id": 1})
        assert view.permissions

        mocked_filter.assert_called_once_with(
            Q(content_type=1)
            & ~Q(codename__icontains="administrate")
            & ~Q(codename__icontains="import")
            & ~Q(codename__icontains="export")
        )

    def test_linked_object(self, mocker, request):
        mocked_contrib_type = mocker.patch("wbcore.contrib.guardian.viewsets.viewsets.ContentType.objects")
        mocked_get_object_for_this_type = mocker.Mock()
        mocked_contrib_type.get.return_value = mocked_get_object_for_this_type
        view = PivotUserObjectPermissionModelViewSet(request=request, kwargs={"content_type_id": 1, "object_pk": 1})
        assert view.linked_object

        mocked_get_object_for_this_type.get_object_for_this_type.assert_called_once()

    @pytest.mark.django_db
    def test_queryset(self, user, request):
        assign_perm("view_user", user, user)
        view = PivotUserObjectPermissionModelViewSet(
            request=request, kwargs={"content_type_id": ContentType.objects.get_for_model(user), "object_pk": user.id}
        )

        permissions = view.permissions
        instance = view.get_queryset().order_by("id").first()

        for permission in permissions:
            assert permission.codename in instance.keys()

    def test_serializer_class(self, request, mocker):
        view = PivotUserObjectPermissionModelViewSet(request=request, kwargs={"content_type_id": 1, "object_pk": 1})
        mocker.patch.object(view, "permissions", [Permission(codename="view_user")])
        serializer_class = view.get_serializer_class()
        assert "view_user" in serializer_class.Meta.fields  # type: ignore



from rest_framework.reverse import reverse

from wbcore import serializers as wb_serializers

from .models import Currency, CurrencyFXRates


class CurrencyRepresentationSerializer(wb_serializers.RepresentationSerializer):
    """Representation Serializer for Currencies"""

    name_repr = wb_serializers.SerializerMethodField(read_only=True)
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:currency:currency-detail")

    def get_name_repr(self, obj):
        return str(obj)

    class Meta:
        model = Currency
        fields = ("id", "name_repr", "key", "_detail")


class CurrencyModelSerializer(wb_serializers.ModelSerializer):
    """Model Serializer for Currencies"""

    rates_sparkline = wb_serializers.SparklineField(
        x_data_label="fx_rates_date", y_data_label="fx_rates_value", label="Rates"
    )

    @wb_serializers.register_resource()
    def currency_currencyfxrates(self, instance, request, user):
        return {
            "currency_currencyfxrates": reverse(
                "wbcore:currency:currency-currencyfxrates-list", args=[instance.id], request=request
            )
        }

    class Meta:
        model = Currency
        fields = ("id", "title", "symbol", "key", "rates_sparkline", "_additional_resources")


class CurrencyFXRatesModelSerializer(wb_serializers.ModelSerializer):
    class Meta:
        model = CurrencyFXRates
        fields = ("id", "date", "currency", "value")



from rest_framework.reverse import reverse

from wbcore.metadata.configs.endpoints import EndpointViewConfig


class DocumentEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs):
        if "content_type" in self.view.kwargs and "content_id" in self.view.kwargs:
            return reverse(
                "wbcore:documents:document_content_object",
                args=[self.view.kwargs["content_type"], self.view.kwargs["content_id"]],
                request=self.request,
            )
        return super().get_create_endpoint(**kwargs)

    def get_delete_endpoint(self, **kwargs):
        if "pk" in self.view.kwargs:
            if self.view.get_object().system_created:
                return None
        return super().get_delete_endpoint(**kwargs)



from django.utils.translation import gettext as _

from wbcore.metadata.configs.titles import TitleViewConfig


class TeamResultsTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Team Results")


class TeamResultsLeagueTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Table")



from django.urls import include, path

from wbcore.routers import WBCoreRouter

from . import viewsets

router = WBCoreRouter()

# Representation viewsets
router.register(r"sourcerepresentation", viewsets.SourceRepresentationViewSet, basename="sourcerepresentation")
router.register(
    r"importsourcerepresentation", viewsets.ImportSourceRepresentationViewSet, basename="importsourcerepresentation"
)
router.register(
    r"parserhandlerrepresentation", viewsets.ParserHandlerRepresentationViewSet, basename="parserhandlerrepresentation"
)
router.register(
    r"databackendrepresentation", viewsets.DataBackendRepresentationViewSet, basename="databackendrepresentation"
)

# Models viewsets
router.register(r"parserhandler", viewsets.ParserHandlerModelViewSet, basename="parserhandler")
router.register(r"importsource", viewsets.ImportSourceModelViewSet, basename="importsource")
router.register(r"source", viewsets.SourceModelViewSet, basename="source")


urlpatterns = [
    path("", include(router.urls)),
]



import json
from typing import Callable

from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APIRequestFactory
from termcolor import colored

from wbcore.messages import InMemoryMessageStorage
from wbcore.signals import add_additional_resource

from .signals import get_custom_factory
from .utils import (
    get_data_from_factory,
    get_kwargs,
    get_model_factory,
    get_or_create_superuser,
)


class TestModel:
    def __init__(self, _model):
        self.model = _model
        self.factory = get_model_factory(self.model)

    def _raise_warning(self, name_test):
        model_name = self.model.__name__ if self.model else None
        print(  # noqa: T201
            f"- {self.__class__.__name__}:{name_test}",
            colored(
                f"WARNING - {model_name} has no attribute {name_test}",
                "yellow",
            ),
        )

    def _successful_test(self, name_test):
        print(f"- {self.__class__.__name__}:{name_test}", colored("PASSED", "green"))  # noqa: T201

    def test_endpoint_basename(self):
        if not hasattr(self.model, "get_endpoint_basename"):
            self._raise_warning("get_endpoint_basename")
        else:
            assert isinstance(self.model.get_endpoint_basename, Callable), "endpoint basename should be Callable"
            assert reverse(
                f"{self.model.get_endpoint_basename()}-list", request=APIRequestFactory().get("")
            ), "endpoint basename should not be None"
            self._successful_test("test_endpoint_basename")

    def test_representation_endpoint(self):
        if not hasattr(self.model, "get_representation_endpoint"):
            self._raise_warning("get_representation_endpoint")
        else:
            assert isinstance(
                self.model.get_representation_endpoint, Callable
            ), "representation endpoint should be Callable"
            if not self.model.get_representation_endpoint():
                print(  # noqa: T201
                    f"- {self.__class__.__name__}:get_representation_endpoint",
                    colored("ERROR - representation_value_endpoint should not be None. ", "red"),
                    colored("WARNING - representation_value_endpoint should not be None", "yellow"),
                )
            else:
                assert reverse(
                    f"{self.model.get_representation_endpoint()}", request=APIRequestFactory().get("")
                ), "representation endpoint should not be None"
                self._successful_test("test_representation_endpoint")

    def test_representation_value_key(self):
        if not hasattr(self.model, "get_representation_value_key"):
            self._raise_warning("get_representation_value_key")
        else:
            assert self.model.get_representation_value_key(), "representation_value_key should not be None"
            self._successful_test("test_representation_value_key")

    def test_representation_label_key(self):
        if not hasattr(self.model, "get_representation_label_key"):
            self._raise_warning("get_representation_label_key")
        else:
            assert self.model.get_representation_label_key(), "representation_label_key should not be None"
            self._successful_test("test_representation_label_key")

    def test_str(self):
        if self.factory is None:
            print(  # noqa: T201
                f"- {self.__class__.__name__}:test_str",
                colored(f"WARNING - factory not found for {self.model.__name__ if self.model else None}", "yellow"),
            )
        else:
            obj = self.factory()
            assert str(obj), "str() function should return a string"
            assert isinstance(str(obj), str), "str() function should return a string"
            self._successful_test("test_str")

    def execute_test(self):
        print("\n")  # noqa: T201
        self.test_endpoint_basename()
        self.test_representation_endpoint()
        self.test_representation_value_key()
        self.test_representation_label_key()
        self.test_str()


class TestSerializer:
    def __init__(self, _serializer):
        self.serializer = _serializer
        self.factory = get_model_factory(self.serializer.Meta.model) if self.serializer else None

    def test_serializer(self):
        if self.factory is None:
            print(  # noqa: T201
                f"\n- {self.__class__.__name__}:test_serializer: {self.serializer.__name__ if self.serializer else None}",
                colored(
                    "WARNING - factory not found for " + self.serializer.Meta.model.__name__
                    if self.serializer
                    else None,
                    "yellow",
                ),
            )
        else:
            request = APIRequestFactory().get("")
            request.user = get_or_create_superuser()
            request.parser_context = {}
            serializer = self.serializer(self.factory(), context={"request": request})
            assert serializer.data, str(serializer.data) + " should not be empty"
            print(  # noqa: T201
                f"\n- {self.__class__.__name__}:test_serializer {self.serializer.__name__ if self.serializer else None}",
                colored("PASSED", "green"),
            )

    def test_additional_resources(self):
        if self.factory is None:
            print(  # noqa: T201
                f"\n- {self.__class__.__name__}:test_serializer: {self.serializer.__name__ if self.serializer else None}",
                colored(
                    "WARNING - factory not found for " + self.serializer.Meta.model.__name__
                    if self.serializer
                    else None,
                    "yellow",
                ),
            )
        else:
            request = APIRequestFactory().get("")
            request.user = get_or_create_superuser()
            request.parser_context = {}
            instance = self.factory()
            self.serializer(instance, context={"request": request})
            add_additional_resource.send(
                sender=self.serializer.__class__,
                serializer=self.serializer,
                instance=instance,
                request=request,
                user=request.user,
            )
            # TODO

    def execute_test(self):
        self.test_serializer()
        # self.test_additional_resources()


class ParentViewset:
    def __init__(self, mvs, factory=None):
        self.mvs = mvs
        request = request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        # self.model = mvs(kwargs=dict(), request=request).get_serializer_class().Meta.model if mvs else None
        self.model = mvs.get_model() if mvs else None
        self.factory = (
            factory
            if factory
            else custom_factory[0][1]
            if (custom_factory := get_custom_factory.send(self.mvs))
            else get_model_factory(self.model)
        )

    def _get_mixins_data(self, type="GET", dump_data=False, data=None):  # noqa: C901
        api_request = APIRequestFactory()
        superuser = get_or_create_superuser()
        kwargs = None
        obj = superuser if self.factory._meta.model.__name__ == "User" else self.factory()
        if type == "GET":
            request = api_request.get("")
        elif type == "OPTIONS":
            request = api_request.options("")
        elif type == "DELETE":
            request = api_request.delete("")
        elif type == "POST":
            if not data:
                data = get_data_from_factory(obj, self.mvs, superuser=superuser, delete=True, factory=self.factory)
            if dump_data:
                data = json.dumps(data)
                request = api_request.post("", data=data, content_type="application/json")
            else:
                try:
                    request = api_request.post("", data)
                except Exception:
                    try:
                        request = api_request.post("", data, format="json")
                    except TypeError:
                        try:
                            data = json.dumps(data)
                            request = api_request.post("", data=data, content_type="application/json")
                        except TypeError as e:
                            raise e
            request.user = superuser
            kwargs = get_kwargs(obj, self.mvs, request=request, data=data)
        else:  # "UPDATE", "PATCH"
            if not data:
                data = get_data_from_factory(obj, self.mvs, superuser=superuser, update=True)
            if type == "UPDATE":
                try:
                    request = api_request.put("", data)
                except Exception:
                    try:
                        request = api_request.put("", data, format="json")
                    except TypeError as e:
                        raise e
                if dump_data:
                    data = json.dumps(data)
                    request = api_request.put("", data=data, content_type="application/json")
            else:  # PATCH
                try:
                    request = api_request.patch("", data)
                except Exception:
                    try:
                        request = api_request.patch("", data, format="json")
                    except TypeError as e:
                        raise e
                if dump_data:
                    data = json.dumps(data)
                    request = api_request.patch("", data=data, content_type="application/json")
            request.user = superuser
            kwargs = get_kwargs(obj, self.mvs, request=request, data=data)
        request.user = superuser
        request.query_params = dict()
        kwargs = get_kwargs(obj, self.mvs, request) if kwargs is None else kwargs
        request._messages = InMemoryMessageStorage(request)
        return obj, request, kwargs, data

    def _raise_warning_factory(self, name_test):
        print(  # noqa: T201
            f"\n- {self.__class__.__name__}:{name_test}",
            colored(
                f"WARNING - factory not found for {self.model.__name__ if self.model else None}",
                "yellow",
            ),
        )

    def _get_endpoint_config(self, request, kwargs, obj=None):
        self.mvs.kwargs = kwargs
        if obj:
            self.mvs.kwargs.update({"pk": obj.pk})
        ep = self.mvs(request=request).endpoint_config_class(
            self.mvs(request=request), request=request, instance=False
        )
        if obj:
            ep.instance = obj
        return ep


class TestRepresentationViewSet(ParentViewset):
    # Test "get": "list"
    def test_list_representation_viewset(self):
        if self.factory is None:
            self._raise_warning_factory("test_list_representation_viewset")
        else:
            _, request, kwargs, _ = self._get_mixins_data()
            vs = self.mvs.as_view({"get": "list"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK, str(response.status_code) + f" == 200 ({response.data})"
            assert response.data, str(response.data) + " should not be empty"
            print(  # noqa: T201
                f"\n- {self.__class__.__name__}:test_list_representation_viewset", colored("PASSED", "green")
            )

    # Test "get": "retrieve"
    def test_instance_representation_viewset(self):
        if self.factory is None:
            self._raise_warning_factory("test_instance_representation_viewset")
        else:
            obj, request, kwargs, _ = self._get_mixins_data()
            vs = self.mvs.as_view({"get": "retrieve"})
            response = vs(request, **kwargs, pk=obj.pk)
            assert response.status_code == status.HTTP_200_OK, str(response.status_code) + f" == 200 ({response.data})"
            assert response.data.get("instance"), str(response.data.get("instance")) + " should not be empty"
            print(  # noqa: T201
                f"- {self.__class__.__name__}:test_instance_representation_viewset", colored("PASSED", "green")
            )

    def execute_test(self):
        self.test_list_representation_viewset()
        self.test_instance_representation_viewset()


class TestViewSet(ParentViewset):
    # test viewset Option request
    def test_option_request(self, is_instance=False):
        if self.factory is None:
            self._raise_warning_factory("test_option_request")
        else:
            obj, request, kwargs, _ = self._get_mixins_data("OPTIONS")
            if not is_instance:
                ep = self._get_endpoint_config(request, kwargs)
            else:
                ep = self._get_endpoint_config(request, kwargs, obj)
                if ep.get_instance_endpoint():
                    kwargs["pk"] = obj.pk
            vs = self.mvs.as_view({"options": "options"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK, str(response.status_code) + f" == 200 ({response.data})"
            assert response.data, str(response.data) + " should not be empty"
            if "buttons" in response.data.keys():
                if "custom_instance" in response.data.get("buttons").keys():
                    assert (
                        list(response.data["buttons"]["custom_instance"])
                        or len(list(response.data["buttons"]["custom_instance"])) == 0
                    )
            assert response.data.get("fields"), str(response.data.get("fields")) + " should not be None"
            assert response.data.get("identifier"), str(response.data.get("identifier")) + " should not be None"
            # assert response.data.get("pagination")
            # assert response.data.get("pk")
            # assert response.data.get("type")
            # assert response.data.get("filterset_fields")
            # assert response.data.get("search_fields")
            # assert response.data.get("ordering_fields")
            # assert response.data.get("buttons"), str(response.data.get("buttons")) + " should not be None"  # TODO: Refactor - buttons can be none
            assert response.data.get("display"), str(response.data.get("display")) + " should not be None"
            assert response.data.get("titles"), str(response.data.get("titles")) + " should not be None"
            assert response.data.get("endpoints"), str(response.data.get("endpoints")) + " should not be None"
            # assert response.data.get("preview")
            print(f"- {self.__class__.__name__}:test_option_request", colored("PASSED", "green"))  # noqa: T201

    # ----- LIST ROUTE TEST ----- #
    # Test viewset "get": "list"
    def test_get_request(self):
        if self.factory is None:
            self._raise_warning_factory("test_get_request")
        else:
            _, request, kwargs, _ = self._get_mixins_data("GET")
            vs = self.mvs.as_view({"get": "list"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK, str(response.status_code) + f" == 200 ({response.data})"
            assert response.data, str(response.data) + " should not be empty"
            print(f"- {self.__class__.__name__}:test_get_request", colored("PASSED", "green"))  # noqa: T201

    # Test viewset "get": "list" -> aggregation
    def test_aggregation(self, name_field=None):
        if self.factory is None:
            self._raise_warning_factory("test_aggregation")
        else:
            _, request, kwargs, _ = self._get_mixins_data("GET")

            vs = self.mvs.as_view({"get": "list"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK, str(response.status_code) + f" == 200 ({response.data})"
            assert response.data, str(response.data) + " should not be empty"
            if not response.data.get("aggregates"):
                print(  # noqa: T201
                    f"- {self.__class__.__name__}:test_aggregation:" + self.mvs.__name__,
                    colored("WARNING - aggregates not found in " + self.mvs.__name__, "yellow"),
                )
            else:
                if name_field:
                    assert response.data.get("aggregates").get(name_field), name_field + "not found in aggregates"
            print(f"- {self.__class__.__name__}:test_aggregation", colored("PASSED", "green"))  # noqa: T201

    # Test viewset "get": "list" with client and endpoint
    def test_get_endpoint(self, client):
        if self.factory is None:
            self._raise_warning_factory("test_get_endpoint")
        else:
            _, request, kwargs, _ = self._get_mixins_data("GET")
            ep = self._get_endpoint_config(request, kwargs)
            client.force_login(request.user)
            if ep.get_endpoint():
                response = client.get(ep.get_endpoint())
            else:
                response = client.get(ep._get_list_endpoint())
            assert response.status_code == status.HTTP_200_OK, str(response.status_code) + f" == 200 ({response.data})"
            assert response.data, str(response.data) + " should not be empty"
            print(f"- {self.__class__.__name__}:test_get_endpoint", colored("PASSED", "green"))  # noqa: T201

    # Test viewset "post": "create"
    def test_post_request(self, dump_data=False, data=None):
        if self.factory is None:
            self._raise_warning_factory("test_post_request")
        else:
            _, request, kwargs, _ = self._get_mixins_data("POST", dump_data=dump_data, data=data)
            vs = self.mvs.as_view({"post": "create"})
            ep = self._get_endpoint_config(request, kwargs)
            ep_create = ep.get_create_endpoint()
            response = vs(request, **kwargs)
            if ep_create:
                if response.status_code == status.HTTP_400_BAD_REQUEST:
                    try:
                        _, request, kwargs, _ = self._get_mixins_data(type="POST", dump_data=True, data=data)
                        vs = self.mvs.as_view({"post": "create"})
                        response2 = vs(request, **kwargs)
                        assert response2.status_code == status.HTTP_201_CREATED, (
                            str(response2.status_code)
                            + f" == 201 ({response2.data}). Result of first attempt[without dump_data]({response.data})"
                        )
                        response = response2
                    except TypeError:
                        assert response.status_code == status.HTTP_201_CREATED, (
                            str(response.status_code) + f" == 201 ({response.data})"
                        )
                else:
                    assert response.status_code == status.HTTP_201_CREATED, (
                        str(response.status_code) + f" == 201 ({response.data})"
                    )
                assert response.data.get("instance"), str(response.data.get("instance")) + " should not be empty"
            else:
                assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED, (
                    str(response.status_code) + f" == 405 ({response.data})"
                )
            print(f"- {self.__class__.__name__}:test_post_request", colored("PASSED", "green"))  # noqa: T201

    # Test viewset "post": "create" with client and endpoint
    def test_post_endpoint(self, client, dump_data=False, data=None):
        if self.factory is None:
            self._raise_warning_factory("test_post_endpoint")
        else:
            _, request, kwargs, data = self._get_mixins_data("POST", dump_data=dump_data, data=data)
            ep = self._get_endpoint_config(request, kwargs)
            ep_create = ep.get_create_endpoint()
            client.force_login(request.user)
            response = client.post(ep_create, data)
            if ep_create:
                if response.status_code == status.HTTP_400_BAD_REQUEST:
                    try:
                        _, request, kwargs, data = self._get_mixins_data(type="POST", dump_data=True, data=data)
                        ep = self._get_endpoint_config(request, kwargs)
                        response2 = client.post(ep_create, data)
                        assert response2.status_code == status.HTTP_201_CREATED, (
                            str(response2.status_code)
                            + f" == 201 ({response2.data}). Result of first attempt[without dump_data]({response.data})"
                        )
                        response = response2
                    except TypeError:
                        assert response.status_code == status.HTTP_201_CREATED, (
                            str(response.status_code) + f" == 201 ({response.data})"
                        )
                else:
                    assert response.status_code == status.HTTP_201_CREATED, (
                        str(response.status_code) + f" == 201 ({response.data})"
                    )
                assert response.data.get("instance"), str(response.data.get("instance")) + " should not be empty"
            else:
                assert response.status_code == status.HTTP_404_NOT_FOUND, (
                    str(response.status_code) + f" == 404 ({response.data})"
                )
            print(f"- {self.__class__.__name__}:test_post_endpoint", colored("PASSED", "green"))  # noqa: T201

    # Test viewset "delete": "destroy_multiple"
    def test_destroy_multiple(self):
        if self.factory is None:
            self._raise_warning_factory("test_destroy_multiple")
        else:
            _, request, kwargs, _ = self._get_mixins_data("DELETE")
            for _ in range(4):
                self.factory()
            vs = self.mvs.as_view({"delete": "destroy_multiple"})
            ep = self._get_endpoint_config(request, kwargs)
            ep_delete = ep.get_delete_endpoint()
            response = vs(request, **kwargs)
            if ep_delete:
                assert response.status_code == status.HTTP_204_NO_CONTENT, (
                    str(response.status_code) + f" == 204 ({response.data})"
                )
            else:
                assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED, (
                    str(response.status_code) + f" == 405 ({response.data})"
                )
            print(f"- {self.__class__.__name__}:test_destroy_multiple", colored("PASSED", "green"))  # noqa: T201

    # -------------- DETAIL ROUTE TEST ------------------#
    # Test viewset "get": "retrieve"
    def test_retrieve_request(self):
        if self.factory is None:
            self._raise_warning_factory("test_retrieve_request")
        else:
            obj, request, kwargs, _ = self._get_mixins_data("GET")
            vs = self.mvs.as_view({"get": "retrieve"})
            ep = self._get_endpoint_config(request, kwargs, obj)
            response = vs(request, **kwargs)
            if ep._get_instance_endpoint():
                assert response.status_code == status.HTTP_200_OK, (
                    str(response.status_code) + f" == 200 ({response.data})"
                )
                assert response.data.get("instance"), str(response.data.get("instance")) + " should not be empty"
            print(f"- {self.__class__.__name__}:test_retrieve_request", colored("PASSED", "green"))  # noqa: T201

    # Test "delete": "destroy"
    def test_delete_request(self):
        if self.factory is None:
            self._raise_warning_factory("test_delete_request")
        else:
            obj, request, kwargs, _ = self._get_mixins_data("DELETE")
            vs = self.mvs.as_view({"delete": "destroy"})
            ep = self._get_endpoint_config(request, kwargs, obj)
            ep_delete = ep.get_delete_endpoint()
            response = vs(request, **kwargs)
            if ep_delete:
                assert response.status_code == status.HTTP_204_NO_CONTENT, (
                    str(response.status_code) + f" == 204 ({response.data})"
                )
            else:
                assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED, (
                    str(response.status_code) + f" == 405 ({response.data})"
                )
            print(f"- {self.__class__.__name__}:test_delete_request", colored("PASSED", "green"))  # noqa: T201

    # Test "put": "update"
    def test_update_request(self, dump_data=False, data=None):
        if self.factory is None:
            self._raise_warning_factory("test_update_request")
        else:
            obj, request, kwargs, _ = self._get_mixins_data("UPDATE", dump_data=dump_data, data=data)
            vs = self.mvs.as_view({"put": "update"})
            ep = self._get_endpoint_config(request, kwargs, obj)
            ep_update = ep.get_update_endpoint()
            response = vs(request, **kwargs)
            if ep_update:
                if response.status_code == status.HTTP_400_BAD_REQUEST:
                    try:
                        obj, request, kwargs, _ = self._get_mixins_data(type="UPDATE", dump_data=True, data=data)
                        vs = self.mvs.as_view({"put": "update"})
                        ep = self._get_endpoint_config(request, kwargs, obj)
                        response2 = vs(request, **kwargs)
                        assert response2.status_code == status.HTTP_200_OK, (
                            str(response2.status_code)
                            + f" == 200 ({response2.data}). Result of first attempt[without dump_data]({response.data})"
                        )
                        response = response2
                    except TypeError:
                        assert response.status_code == status.HTTP_200_OK, (
                            str(response.status_code) + f" == 200 ({response.data})"
                        )
                else:
                    assert response.status_code == status.HTTP_200_OK, (
                        str(response.status_code) + f" == 200 ({response.data})"
                    )
                assert response.data.get("instance"), str(response.data.get("instance")) + " should not be empty"
            else:
                assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED, (
                    str(response.status_code) + f" == 405 ({response.data})"
                )
            print(f"- {self.__class__.__name__}:test_update_request", colored("PASSED", "green"))  # noqa: T201

    # Test "patch": "partial_update",
    def test_patch_request(self, dump_data=False, data=None):
        if self.factory is None:
            self._raise_warning_factory("test_patch_request")
        else:
            obj, request, kwargs, data = self._get_mixins_data("PATCH", dump_data=dump_data, data=data)
            vs = self.mvs.as_view({"patch": "partial_update"})
            ep = self._get_endpoint_config(request, kwargs, obj)
            ep_instance = ep.get_instance_endpoint()
            ep_update = ep.get_update_endpoint()
            response = vs(request, **kwargs, data=data)
            if ep_instance and ep_update:
                if response.status_code == status.HTTP_400_BAD_REQUEST:
                    try:
                        obj, request, kwargs, data = self._get_mixins_data(type="PATCH", dump_data=True, data=data)
                        vs = self.mvs.as_view({"patch": "partial_update"})
                        ep = self._get_endpoint_config(request, kwargs, obj)
                        response2 = vs(request, **kwargs, data=data)
                        assert response2.status_code == status.HTTP_200_OK, (
                            str(response2.status_code)
                            + f" == 200 ({response2.data}). Result of first attempt[without dump_data]({response.data})"
                        )
                        response = response2
                    except TypeError:
                        assert response.status_code == status.HTTP_200_OK, (
                            str(response.status_code) + f" == 200 ({response.data})"
                        )
                else:
                    assert response.status_code == status.HTTP_200_OK, (
                        str(response.status_code) + f" == 200 ({response.data})"
                    )
                assert response.data.get("instance"), str(response.data.get("instance")) + " should not be empty"
            else:
                assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED, (
                    str(response.status_code) + f" == 405 ({response.data})"
                )
            print(f"- {self.__class__.__name__}:test_patch_request", colored("PASSED", "green"))  # noqa: T201

    # ----- LIST ROUTE TEST ----- #
    def execute_test_list_endpoint(self, client, aggregates=None):
        print("\n")  # noqa: T201
        self.test_option_request()
        self.test_get_request()
        self.test_aggregation(aggregates)
        # self.test_get_endpoint(client)
        self.test_post_request()
        # self.test_post_endpoint(client)
        self.test_destroy_multiple()

    # ----- DETAIL ROUTE TEST ----- #
    def execute_test_detail_endpoint(self):
        self.test_option_request(is_instance=True)
        self.test_retrieve_request()
        self.test_update_request()
        self.test_patch_request()
        self.test_delete_request()


class TestPandasView(TestViewSet):
    def test_get_request(self):
        if self.factory is None:
            self._raise_warning_factory("test_get_request")
        else:
            _, request, kwargs, _ = self._get_mixins_data("GET")
            vs = self.mvs.as_view({"get": "list"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK, str(response.status_code) + f" == 200 ({response.data})"
            assert response.data, str(response.data) + " should not be empty"
            print(f"- {self.__class__.__name__}:test_get_request", colored("PASSED", "green"))  # noqa: T201

    def execute_test(self):
        print("\n")  # noqa: T201
        self.test_get_request()


class TestChartViewSet(TestViewSet):
    def test_option_request(self, is_instance=False):
        if self.factory is None:
            self._raise_warning_factory("test_option_request")
        else:
            obj, request, kwargs, _ = self._get_mixins_data("OPTIONS")
            if not is_instance:
                ep = self._get_endpoint_config(request, kwargs)
            else:
                ep = self._get_endpoint_config(request, kwargs, obj)
                if ep.get_instance_endpoint():
                    kwargs["pk"] = obj.pk

            vs = self.mvs.as_view({"options": "options"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK, str(response.status_code) + f" == 200 ({response.data})"
            assert response.data, str(response.data) + " should not be empty"
            if "buttons" in response.data.keys():
                if "custom_instance" in response.data.get("buttons").keys():
                    assert (
                        list(response.data["buttons"]["custom_instance"])
                        or len(list(response.data["buttons"]["custom_instance"])) == 0
                    )
            assert response.data.get("identifier"), str(response.data.get("identifier")) + " should not be None"
            assert response.data.get("type") == "chart", "type of view should be chart"
            assert response.data.get("buttons"), str(response.data.get("buttons")) + " should not be None"
            assert response.data.get("display"), str(response.data.get("display")) + " should not be None"
            assert response.data.get("titles"), str(response.data.get("titles")) + " should not be None"
            assert response.data.get("endpoints"), str(response.data.get("endpoints")) + " should not be None"
            print(f"- {self.__class__.__name__}:test_option_request", colored("PASSED", "green"))  # noqa: T201

    def execute_test(self):
        print("\n")  # noqa: T201
        self.test_option_request()
        self.test_get_request()



from typing import TYPE_CHECKING

from wbcore.metadata.configs.titles import TitleViewConfig

if TYPE_CHECKING:
    from wbcore.contrib.guardian.viewsets import PivotUserObjectPermissionModelViewSet


class PivotUserObjectPermissionTitleViewConfig(TitleViewConfig):
    view: "PivotUserObjectPermissionModelViewSet"

    def get_list_title(self) -> str:
        return f"Permissions for {self.view.linked_object}"

    def get_create_title(self) -> str:
        return f"Add permissions for {self.view.linked_object}"



from datetime import datetime

from django.contrib.contenttypes.models import ContentType
from django.db import models

from .imports import ImportExportHandler
from .models import Source


class ImportMixin(models.Model):
    import_export_handler_class: type[ImportExportHandler] | None = None
    import_source = models.ForeignKey(
        "io.ImportSource",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )

    @classmethod
    def import_data(cls, execution_time: datetime | None = None, synchronous: bool = False, **kwargs):
        """
        Allow Resource to trigger the workflow from the instance
        Args:
            execution_time: The time at which the task needs to be trigger
            **kwargs: keyword arguments passed down the workflow
        """
        content_type = ContentType.objects.get_for_model(cls)
        model_name = f"{content_type.app_label}.{content_type.model}"
        for source in Source.objects.filter(
            parser_handler__handler__iexact=model_name, data_backend__passive_only=False, is_active=True
        ):
            source.trigger_workflow(
                execution_time=execution_time, synchronous=synchronous, only_handler=model_name, **kwargs
            )

    class Meta:
        abstract = True



from django.apps import apps
from markdown.extensions import Extension
from markdown.preprocessors import Preprocessor


class FSMPreprocessor(Preprocessor):
    def __init__(self, *args, **kwargs):
        self.indicator = "!!!wbcore-fsm"
        super().__init__(*args, **kwargs)

    def run(self, lines):
        for line in lines:
            if self.indicator in line:
                _, app_model, model_field = line.split(":")
                model = apps.get_model(app_model)

                yield "~~~mermaid\ngraph TD"

                transitions = getattr(model(), f"get_all_{model_field}_transitions")
                for transition in transitions():
                    yield f"\t({transition.source}) --> |{transition.name}|({transition.target})"

                yield "~~~"

            else:
                yield line


class FSMExtension(Extension):
    def extendMarkdown(self, md, md_globals):  # noqa
        md.preprocessors.register(FSMPreprocessor(), "wbcore-fsm", 100)



import logging
from contextlib import suppress
from typing import Any, ClassVar, Type

from django.db.models.fields import NOT_PROVIDED
from rest_framework.fields import empty
from rest_framework.serializers import Field

logger = logging.getLogger(__name__)


def decorator(position: str, value: str, decorator_type: str = "icon") -> dict:
    if position not in ("left", "right"):
        raise ValueError("Decorator Position can only be right or left")
    if decorator_type not in ("icon", "text"):
        raise ValueError("Decorator Type can only be icon or text")
    return {"position": position, "value": value, "type": decorator_type}


class WBCoreSerializerFieldMixin:
    """
    A mixin that takes care of adding all the necessary magic to each implementation
    of the serializer fields
    """

    def __init__(
        self,
        *args,
        decorators=None,
        depends_on=None,
        math=None,
        extra=None,
        read_only=False,
        copyable=None,
        related_key=None,
        on_unsatisfied_deps="read_only",
        clear_dependent_fields=True,
        **kwargs,
    ):
        if not decorators:
            decorators = list()
        if not depends_on:
            depends_on = list()
        self.extra = extra
        self.decorators = decorators
        self.depends_on = depends_on
        self.math = math
        self.copyable = copyable
        # To satisfy DRF assertion, we detect if the readonly given is a callable and in that case, we consider the default read only attribute to be true
        self._callable_read_only = None
        if callable(read_only):
            self._callable_read_only = read_only
            read_only = True
        self.related_key = related_key
        self.on_unsatisfied_deps = on_unsatisfied_deps
        self.clear_dependent_fields = clear_dependent_fields
        super().__init__(*args, read_only=read_only, **kwargs)

    def _evaluate_read_only(self, field_name, parent):
        if self._callable_read_only:
            # if view is present, we use it as parameter to the callable, otherwise we assume read_only is True
            if view := self.view:
                self.read_only = self._callable_read_only(view)
            else:
                self.read_only = True

        if meta := getattr(parent, "Meta", None):
            read_only_fields = getattr(meta, "read_only_fields", [])
            # By default, DRF do not apply extra_kwargs to declared field, this leds to confusion in our codebase so we take the conscious decision to make read_only_fields take presence over declared field attribute
            if field_name in read_only_fields:
                self.read_only = True

    def bind(self, field_name, parent):
        super().bind(field_name, parent)
        # After binding to the parent happens, we evaluate the callable readonly if given
        self._evaluate_read_only(field_name, parent)

    def get_related_key(self) -> str:
        return self.related_key or self.source

    @property
    def view(self):
        # Helper property to return the stored instantiated view in the context
        return self.context.get(
            "view",
            getattr(self.context.get("request"), "parser_context", {}).get("view"),
        )

    def set_attr(self, attr, value):
        setattr(self, attr, value)
        self._kwargs[attr] = value

    def append_depend_on(self, depend_on):
        existing_fields = [elem["field"] for elem in self.depends_on]
        if depend_on["field"] not in existing_fields:
            self.depends_on.append(depend_on)
            self._kwargs["depends_on"] = self.depends_on

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        if meta := getattr(self.parent, "Meta", None):
            if field_name in getattr(meta, "required_fields", []):
                self.required = True

        representation = {
            "key": field_name,
            "label": getattr(self, "label", None),
            "type": getattr(self, "field_type", "undefined"),
            "required": getattr(self, "required", True),
            "read_only": getattr(self, "read_only", False),
        }

        default = getattr(self, "default", None)
        if default is None or default == empty or default == NOT_PROVIDED:
            with suppress(Exception):  # TODO Add some explicit exception handling
                default = self.parent.Meta.model._meta._forward_fields_map[field_name].default

        if default is not None and default != empty and default != NOT_PROVIDED:
            if callable(default):
                if getattr(default, "requires_context", False):
                    default = default(self)
                else:
                    default = default()
            if default is not None:
                representation["default"] = self.to_representation(default)

        representation["decorators"] = getattr(self, "decorators", [])
        representation["depends_on"] = getattr(self, "depends_on", [])

        for _attr in ["help_text", "extra"]:
            attr = getattr(self, _attr, None)
            if attr:
                representation[_attr] = attr

        if self.math:
            representation["math"] = self.math
        if self.copyable:
            representation["copyable"] = self.copyable
        if self.on_unsatisfied_deps != "read_only":
            representation["on_unsatisfied_deps"] = self.on_unsatisfied_deps
        if self.clear_dependent_fields is not True:
            representation["clear_dependent_fields"] = self.clear_dependent_fields
        return field_name, representation

    def validate_empty_values(self, data):
        if isinstance(data, str) and data == "null":
            data = None
        return super().validate_empty_values(data)


class RangeMixin(WBCoreSerializerFieldMixin, Field):
    """
    This mixin represents a range in the form of:
    (value1,value2)

    When using this mixin, you must supply the internal_field (usually the postgres wrapper) that converts the range tuple to some datastructure.
    """

    default_error_messages = {
        "lower_gt_upper": "The upper bound must be greater than the lower bound.",
    }
    internal_field: ClassVar[Type]

    def _transform_range(self, lower, upper, inward: bool | None = None):
        return lower, upper

    def to_internal_value(self, data) -> Any:
        # If formdata is used, we need to separate the string into the range
        if isinstance(data, str) and "," in data:
            data = data.split(",")
        lower = upper = None
        if isinstance(data, list):
            if len(data) > 0 and data[0]:
                lower = super().to_internal_value(data[0])
            if len(data) > 1 and data[1]:
                upper = super().to_internal_value(data[1])
        lower, upper = self._transform_range(lower, upper, inward=True)
        if lower and upper and lower >= upper:
            self.fail("lower_gt_upper")
        return self.internal_field(lower, upper)

    def to_representation(self, instance):
        if hasattr(instance, "lower"):
            lower = instance.lower
        else:
            lower = instance[0]

        if hasattr(instance, "upper"):
            upper = instance.upper
        else:
            upper = instance[1]
        lower, upper = self._transform_range(lower, upper, inward=False)
        with suppress(Exception):
            lower = super().to_representation(lower)
        with suppress(Exception):
            upper = super().to_representation(upper)
        return lower, upper



from typing import Callable


def register_config(func: Callable) -> Callable:
    """
    A decorator that wraps a function and adds `_is_config=True` to it. This is needed for the auto-discover
    of all configs. This function should only be used in a module called configs, otherwise the auto-discover
    functionality won't work.
    """
    func._is_config = True
    return func



from wbcore.contrib.authentication.factories import (
    AuthenticatedPersonFactory,
    SuperUserFactory,
)
from wbcore.contrib.authentication.models import User


def create_main_company_user(user_password: str) -> User:
    """Creates a user for the main company.

    Args:
        user_password (str): The password used for the user

    Returns:
        User: The newly created user.
    """
    user = SuperUserFactory(plaintext_password=user_password)
    profile = AuthenticatedPersonFactory(user_account=user)
    user.profile = profile
    return user



from .contacts import (
    AddressContactEntryEndpointConfig,
    BankingContactEndpointConfig,
    BankingContactEntryEndpointConfig,
    EmailContactEntryEndpointConfig,
    SocialMediaContactEntryEndpointConfig,
    TelephoneContactEntryEndpointConfig,
    WebsiteContactEntryEndpointConfig,
)
from .entries import (
    CompanyModelEndpointConfig,
    EntryModelEndpointConfig,
    PersonModelEndpointConfig,
    UserIsManagerEndpointConfig,
)
from .relationships import (
    ClientManagerEndpoint,
    EmployeeEmployerEndpointConfig,
    EmployerEmployeeEndpointConfig,
    RelationshipEntryModelEndpoint,
    UserIsClientEndpointConfig,
)



import mimetypes
import os
from contextlib import suppress
from datetime import datetime
from io import BytesIO
from typing import Generator, Optional

import magic
import requests
from wbcore.contrib.io.backends.abstract import AbstractDataBackend
from wbcore.contrib.io.backends.utils import register


@register("Default Stream", save_data_in_import_source=True)
class DataBackend(AbstractDataBackend):
    """
    DataBackend for stream based import
    """

    def __init__(
        self,
        url: Optional[str] = None,
        **kwargs,
    ):
        """
        Args:
            url: URL to fetch the stream from
            **kwargs:
        Raises:
            ValueError: If url not found in the keywords argument
        """
        if not url:
            raise ValueError("This backend needs a valid url")

        self.url = url

    @classmethod
    def _check_content_type(cls, output: BytesIO, filename: str) -> bool:
        """
        Check if given bytes stream matches the corresponding filename extension

        Args:
            output: The file bytes stream
            filename: The file name (with extension)

        Returns:
            True if the contents matches

        Raises:
            MagicException: If error happens while extracting content type
        """
        try:
            buffer_content_type = magic.from_buffer(output.read(1024), mime=True)
            guess_extensions = mimetypes.guess_extension(buffer_content_type)
            return guess_extensions == os.path.splitext(filename)[1]
        except magic.MagicException:
            return False

    def get_files(
        self,
        execution_time: datetime,
        base_filename: Optional[str] = "default_stream",
        extension: Optional[str] = "xml",
        validate_content_type: Optional[bool] = True,
        **kwargs,
    ) -> Generator[tuple[str, BytesIO], None, None]:
        """
        Connect to file stream and download its content as bytes stream

        Args:
            execution_time: When the import is executed.
            base_filename: The resulting filename. Defaults to "default_stream".
            extension: The resulting file extension. Defaults to ".xml".
            validate_content_type: If true, will sanitize file content with its given extension. Defaults to True.
            **kwargs:

        Returns:
            A generator of tuple of filename and bytes stream

        """
        with suppress(requests.ConnectionError):
            headers = kwargs.get("headers", {})
            r = requests.get(self.url, headers=headers, timeout=10)
            if r.ok and (content := r.content):
                content_file = BytesIO()
                content_file.write(content)
                content_file.seek(0)
                file_name = f"{base_filename}_{datetime.timestamp(execution_time)}.{extension}"
                if not validate_content_type or (
                    validate_content_type and self._check_content_type(content_file, file_name)
                ):
                    yield file_name, content_file



import factory

from ..models import Token


class TokenFactory(factory.django.DjangoModelFactory):
    valid_until = None
    protected_view_name = "wbcore:authentication:user"
    number_usage_left = None
    is_valid = True
    user = factory.SubFactory("wbcore.contrib.authentication.factories.UserFactory")

    class Meta:
        model = Token
        django_get_or_create = ["user", "protected_view_name"]



from django.utils.translation import gettext as _
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import StringPreference

wbcore_section = Section("wbcore")


@global_preferences_registry.register
class DefaultEmptyImagePlaceholderPreference(StringPreference):
    section = wbcore_section
    name = "default_empty_image_placeholder"
    default = "https://stainly-cdn.fra1.cdn.digitaloceanspaces.com/icons/empty_image_placeholder.png"

    verbose_name = _("Default Empty Image Placeholder")
    help_text = _("Default empty image placeholder URL")



from .contacts import (
    AddressContactDisplay,
    AddressContactEntryDisplay,
    BankingContactDisplay,
    BankingContactEntryDisplay,
    EmailContactDisplay,
    EmailContactEntryDisplay,
    SocialMediaContactEntryDisplay,
    TelephoneContactDisplay,
    TelephoneContactEntryDisplay,
    WebsiteContactEntryDisplay,
)
from .entries import (
    BankModelDisplay,
    CompanyModelDisplay,
    EntryModelDisplay,
    PersonModelDisplay,
    UserIsManagerDisplayConfig,
)
from .relationships import (
    ClientManagerModelDisplay,
    ClientManagerRelationshipColor,
    EmployerEmployeeRelationshipDisplayConfig,
    EmployeeEmployerDisplayConfig,
    EmployerEmployeeDisplayConfig,
    RelationshipDisplayConfig,
    RelationshipEntryDisplay,
    RelationshipTypeDisplayConfig,
    UserIsClientDisplayConfig,
)
from .utils import (
    CompanyTypeDisplay,
    CustomerStatusDisplay,
    PositionDisplay,
    SpecializationDisplay,
)



from datetime import date, timedelta

from django.contrib.postgres.aggregates import ArrayAgg
from django.db.models import Q
from rest_framework import filters

from wbcore import viewsets
from wbcore.filters import DjangoFilterBackend

from ..models import Currency, CurrencyFXRates
from ..serializers import CurrencyModelSerializer, CurrencyRepresentationSerializer
from .display import CurrencyDisplayConfig
from .preview import CurrencyPreviewConfig
from .titles import CurrencyTitleConfig


class CurrencyRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Currency.objects.all()
    serializer_class = CurrencyRepresentationSerializer
    filter_backends = (filters.OrderingFilter, filters.SearchFilter)

    ordering = ["name_repr"]
    search_fields = ("name_repr", "key")


class CurrencyModelViewSet(viewsets.ModelViewSet):
    display_config_class = CurrencyDisplayConfig
    title_config_class = CurrencyTitleConfig
    preview_config_class = CurrencyPreviewConfig

    queryset = Currency.objects.all()
    serializer_class = CurrencyModelSerializer

    filter_backends = (DjangoFilterBackend, filters.OrderingFilter)

    filterset_fields = {
        "title": ["exact", "icontains"],
        "symbol": ["exact", "icontains"],
        "key": ["exact", "icontains"],
    }
    ordering_fields = ("title", "symbol", "key")

    ordering = ["title"]

    def get_queryset(self):
        last_rate_date = date.today()
        if CurrencyFXRates.objects.exists():
            last_rate_date = CurrencyFXRates.objects.latest("date").date
        last_rate_date -= timedelta(days=30)
        return (
            super()
            .get_queryset()
            .annotate(
                fx_rates_date=ArrayAgg("fx_rates__date", filter=Q(fx_rates__date__gt=last_rate_date)),
                fx_rates_value=ArrayAgg("fx_rates__value", filter=Q(fx_rates__date__gt=last_rate_date)),
            )
        )



from wbcore import viewsets
from wbcore.contrib.example_app.filters import SportFilter
from wbcore.contrib.example_app.models import Sport
from wbcore.contrib.example_app.serializers import (
    SportModelSerializer,
    SportRepresentationSerializer,
)
from wbcore.contrib.example_app.viewsets.displays import SportDisplayConfig
from wbcore.contrib.example_app.viewsets.titles import SportTitleConfig


class SportModelViewSet(viewsets.ModelViewSet):
    display_config_class = SportDisplayConfig
    queryset = Sport.objects.all()
    search_fields = ("name",)
    serializer_class = SportModelSerializer
    title_config_class = SportTitleConfig
    ordering_fields = ("name", "rules", "match_duration")
    ordering = ("name", "id")
    filterset_class = SportFilter


class SportRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Sport.objects.all()
    serializer_class = SportRepresentationSerializer
    search_fields = ("name",)



from datetime import date

import pytest
from rest_framework.test import APIRequestFactory

from wbcore.utils.date import get_date_interval_from_request, shortcut
from wbcore.utils.date_builder import Day, Now, WeekStart


class TestGetDateInterval:
    def setup_method(self):
        self.request_factory = APIRequestFactory()

    @pytest.mark.parametrize("start_key", ["start", "start_date", "from", "date_gte"])
    @pytest.mark.parametrize("start", [date(2010, 1, 1)])
    @pytest.mark.parametrize("end_key", ["end", "end_date", "to", "date_lte"])
    @pytest.mark.parametrize("end", [date(2010, 1, 1)])
    def test_get_date_interval_from_request(self, start_key, start, end_key, end):
        request = self.request_factory.get(
            path="",
            data={
                start_key: start.strftime("%Y-%m-%d"),
                end_key: end.strftime("%Y-%m-%d"),
            },
        )
        _start, _end = get_date_interval_from_request(request)

        assert _start == start
        assert _end == end

    def test_get_date_interval_from_request_none(self):
        request = self.request_factory.get(
            path="",
        )
        _start, _end = get_date_interval_from_request(request)

        assert _start is None
        assert _end is None

    def test_get_date_interval_from_request_wrong_date_format(self):
        request = self.request_factory.get(path="", data={"start": "abc", "end": "def"})
        _start, _end = get_date_interval_from_request(request)

        assert _start is None
        assert _end is None


class TestShortcut:
    def test_shortcut(self):
        assert shortcut("ABC", WeekStart - Day, Now) == {"label": "ABC", "value": f"{str(WeekStart-Day)},{str(Now)}"}



# Generated by Django 4.1.8 on 2023-05-23 08:13

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("agenda", "0005_building_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="calendaritem",
            name="is_deletable",
            field=models.BooleanField(default=True, verbose_name="Is Deletable"),
        ),
    ]



# Generated by Django 4.1.8 on 2023-04-13 10:33

import uuid

import django.contrib.auth.models
import django.db.models.deletion
import django.utils.timezone
import wbcore.contrib.authentication.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
    ]

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                ("last_login", models.DateTimeField(blank=True, null=True, verbose_name="last login")),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ("email", models.EmailField(max_length=254, unique=True, verbose_name="E-Mail")),
                ("username", models.CharField(max_length=255, unique=True, verbose_name="username")),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Specifies whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=False,
                        help_text="Specifies whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "is_register",
                    models.BooleanField(
                        default=True,
                        help_text="Specifies whether this user has registered its email. ",
                        verbose_name="register",
                    ),
                ),
                (
                    "date_joined",
                    models.DateTimeField(default=django.utils.timezone.now, verbose_name="date joined"),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.permission",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={
                "verbose_name": "user",
                "verbose_name_plural": "users",
                "permissions": [
                    ("administrate_user", "Administrate Users"),
                    ("is_internal_user", "Internal User"),
                ],
                "swappable": "AUTH_USER_MODEL",
            },
            managers=[
                ("objects", wbcore.contrib.authentication.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name="Group",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("auth.group",),
            managers=[
                ("objects", django.contrib.auth.models.GroupManager()),
            ],
        ),
        migrations.CreateModel(
            name="Permission",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("auth.permission",),
            managers=[
                ("objects", django.contrib.auth.models.PermissionManager()),
            ],
        ),
        migrations.CreateModel(
            name="UserActivity",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("IP", models.GenericIPAddressField(blank=True, null=True)),
                ("date", models.DateTimeField(auto_now_add=True)),
                (
                    "status",
                    models.CharField(
                        blank=True,
                        choices=[("Success", "Success"), ("Failed", "Failed")],
                        default="Success",
                        max_length=8,
                        null=True,
                    ),
                ),
                ("jti", models.CharField(blank=True, default="", max_length=32)),
                (
                    "type",
                    models.CharField(
                        blank=True, choices=[("LOGIN", "Login")], default="LOGIN", max_length=16, null=True
                    ),
                ),
                ("user_agent_info", models.CharField(max_length=255)),
                ("latest_refresh", models.DateTimeField(blank=True, null=True)),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="login_activities",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "User Login Activity",
                "verbose_name_plural": "User Login Activities",
            },
        ),
    ]



from django.urls import include, path
from wbcore.contrib.guardian.viewsets import PivotUserObjectPermissionModelViewSet
from wbcore.routers import WBCoreRouter

router = WBCoreRouter()
router.register(
    r"pivoteduserobjectpermission", PivotUserObjectPermissionModelViewSet, basename="pivoteduserobjectpermission"
)

urlpatterns = [
    path("<int:content_type_id>/<int:object_pk>/", include(router.urls)),
]



import factory
from wbcore.contrib.color.models import ColorGradient


class ColorGradientFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ColorGradient

    title = factory.Sequence(lambda n: f"Color Gradient {n}")
    colors = factory.List([factory.Faker("color") for _ in range(10)])



from configurations import values


class I18NL10N:
    MODELTRANS_AVAILABLE_LANGUAGES = ["en", "de", "fr"]
    TIME_ZONE = values.Value("Europe/Berlin", environ_prefix=None)
    LANGUAGE_CODE = values.Value("en-us", environ_prefix=None)
    USE_TZ = values.BooleanValue(True, environ_prefix=None)
    USE_I18N = values.BooleanValue(
        True, environ_prefix=None
    )  # Setting to True lead to a big performance hit when resolver needs to translate url



from django.conf import settings
from django.contrib.sites.models import Site


def base_domain() -> str:
    """A utility method that assembles the current domain. Utilizes the site app from django

    Returns:
        A string containing the current domain as noted in the curret `Site` prefixed with the http scheme

    """
    scheme = "https" if settings.SECURE_SSL_REDIRECT else "http"
    base_domain = Site.objects.get_current().domain
    return f"{scheme}://{base_domain}"


def create_notification_type(
    code: str,
    title: str,
    help_text: str,
    web: bool = True,
    mobile: bool = True,
    email: bool = False,
    resource_button_label: str = "",
    is_lock: bool = False,  # set to true if user cannot modified the preference for this notification type
) -> tuple[str, str, str, bool, bool, bool, str, bool]:
    return (code, title, help_text, web, mobile, email, resource_button_label, is_lock)



from wbcore import viewsets

from ..models import CompanyType, CustomerStatus, Position, Specialization
from ..serializers import (
    CompanyTypeModelSerializer,
    CompanyTypeRepresentationSerializer,
    CustomerStatusModelSerializer,
    CustomerStatusRepresentationSerializer,
    PositionModelSerializer,
    PositionRepresentationSerializer,
    SpecializationModelSerializer,
    SpecializationRepresentationSerializer,
)
from .display import (
    CompanyTypeDisplay,
    CustomerStatusDisplay,
    PositionDisplay,
    SpecializationDisplay,
)
from .titles.utils import (
    CompanyTypeTitleConfig,
    CustomerStatusTitleConfig,
    PositionTitleConfig,
    SpecializationTitleConfig,
)


class CustomerStatusModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/customerstatus.md"
    queryset = CustomerStatus.objects.all()
    serializer_class = CustomerStatusModelSerializer
    display_config_class = CustomerStatusDisplay
    title_config_class = CustomerStatusTitleConfig
    search_fields = ("title",)
    filterset_fields = {
        "title": ["exact", "icontains"],
    }
    ordering_fields = ordering = ("title",)


class CustomerStatusRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = CustomerStatusRepresentationSerializer
    search_fields = ("title",)
    queryset = CustomerStatus.objects.all()


class PositionRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = PositionRepresentationSerializer
    search_fields = ("title",)
    queryset = Position.objects.all()


class PositionModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/position.md"
    queryset = Position.objects.all()
    serializer_class = PositionModelSerializer
    display_config_class = PositionDisplay
    title_config_class = PositionTitleConfig
    search_fields = ("title",)
    filterset_fields = {
        "title": ["exact", "icontains"],
    }
    ordering_fields = ordering = ("title",)


class CompanyTypeModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/companytype.md"
    queryset = CompanyType.objects.all()
    serializer_class = CompanyTypeModelSerializer
    display_config_class = CompanyTypeDisplay
    title_config_class = CompanyTypeTitleConfig
    search_fields = ("title",)
    filterset_fields = {
        "title": ["exact", "icontains"],
    }
    ordering_fields = ordering = ("title",)


class CompanyTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = CompanyTypeRepresentationSerializer
    search_fields = ("title",)
    queryset = CompanyType.objects.all()


class SpecializationModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/specialization.md"
    queryset = Specialization.objects.all()
    serializer_class = SpecializationModelSerializer
    display_config_class = SpecializationDisplay
    title_config_class = SpecializationTitleConfig
    search_fields = ("title",)
    filterset_fields = {
        "title": ["exact", "icontains"],
    }
    ordering_fields = ordering = ("title",)


class SpecializationRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = SpecializationRepresentationSerializer
    search_fields = ("title",)
    queryset = Specialization.objects.all()



from pytest_factoryboy import register
from wbcore.tests.conftest import *

from ..factories import CurrencyFactory, CurrencyFXRatesFactory

register(CurrencyFactory)
register(CurrencyFXRatesFactory)



import logging

from django.contrib.messages import get_messages, warning
from django_fsm import FSMField, Transition, TransitionNotAllowed
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response

from wbcore.messages import InMemoryMessageStorage

logger = logging.getLogger(__name__)


# We have to move the method generation into a method, because we need a real new instance of this method everytime
def get_method(transition, fsm_field_name):
    def method(self, request: Request, transition=transition, pk=None, **kwargs) -> Response:
        return self.fsm_route(request, transition, fsm_field_name, transition.custom.get("serializer", None))

    return method


class FSMViewSetMixinMetaclass(type):
    """Metaclass for dynamically creating all FSM Routes"""

    def __new__(cls, *args, **kwargs):
        _class = super().__new__(cls, *args, **kwargs)

        # The class needs the field FSM_MODELFIELDS to know which transitions it needs to add
        if hasattr(_class, "get_model"):
            model = _class.get_model()

            if model:
                _class.FSM_BUTTONS = getattr(_class, "FSM_BUTTONS", set())
                # The model potentially has multiple FSMFields, which needs to be iterated over
                for field in filter(lambda f: isinstance(f, FSMField), model._meta.fields):
                    # Get all transitions, by calling the partialmethod defined by django-fsm
                    transitions = getattr(model, f"get_all_{field.name}_transitions")(model())

                    # Since the method above can potentially return a transition multiple times
                    # i.e. when a transitions has multiple sources, we need to filter out those transitions
                    _discovered_transitions = list()

                    for transition in transitions:
                        if transition.name in _discovered_transitions:
                            continue
                        else:
                            _discovered_transitions.append(transition.name)

                        # Get the Transition Button and add it to the front of the instance buttons
                        button = transition.custom.get("_transition_button")
                        _class.FSM_BUTTONS.add(button)

                        # Create a method that calls fsm_route with the request and the action name
                        method = get_method(transition, field.name)

                        # We need to manually change the method name, otherwise django-fsm won't
                        # Add this method to the URLs

                        # Wrap the above defined method in the action decorator
                        # IMPORTANT: This needs to happen after we changed the method name
                        # therefore we cannot use the proper decorator
                        method.__name__ = transition.name
                        method.__doc__ = transition.method.__doc__

                        wrapped_method = action(detail=True, methods=["GET", "PATCH"])(method)

                        # Set the method as a attribute of the class that implements this
                        # metaclass
                        setattr(_class, transition.name, wrapped_method)

        return _class


class FSMViewSetMixin(metaclass=FSMViewSetMixinMetaclass):
    def handle_exception(self, exc: Exception) -> Response:
        if isinstance(exc, TransitionNotAllowed):
            return Response({"non_field_errors": str(exc)}, status=status.HTTP_400_BAD_REQUEST)
        return super().handle_exception(exc)

    def fsm_route(
        self, request: Request, transition: Transition, fsm_field_name: str, serializer_class=None
    ) -> Response:
        action = transition.name
        obj = self.get_object()

        if not serializer_class:
            serializer_class = self.get_serializer_class()
        serializer_context = self.get_serializer_context()

        if request.method == "GET":
            return Response(serializer_class(instance=obj, context=serializer_context).data)

        serializer = serializer_class(instance=obj, data=request.data, partial=True, context=serializer_context)
        if serializer.is_valid():
            if len(serializer.validated_data) > 0:
                obj = serializer.save()
        else:
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        errors = None
        if (can_action_method := getattr(obj, f"can_{action}", None)) and callable(can_action_method):
            errors = can_action_method()

        if errors is None or len(errors.keys()) == 0:
            obj.fsm_context = {"current_user": request.user, "request": request}
            if getattr(obj, fsm_field_name) != transition.target:
                if (action_method := getattr(obj, action, None)) and callable(action_method):
                    warnings = action_method(by=request.user)
                    obj.save()
                    if (post_action_method := getattr(obj, f"post_{action}", None)) and callable(post_action_method):
                        post_action_method(by=request.user)
                    # we extend the framework to allow action to successfully return but notify any possible warning. We use the message framework to communicate these warnings to the user
                    if warnings:
                        if isinstance(warnings, list):
                            html = "<ul>" + "".join(f"<li>{e}</li>" for e in warnings) + "</ul>"
                        else:
                            html = "<p>" + warnings + "</p>"
                        warning(request, html, extra_tags="auto_close=0")

            serializer = serializer_class(instance=obj, context=serializer_context)
            storage = get_messages(request._request)
            data = {"instance": serializer.data}
            if isinstance(storage, InMemoryMessageStorage):
                data["messages"] = list(storage.serialize_messages())
            return Response(data)

        return Response(errors, status=status.HTTP_412_PRECONDITION_FAILED)



from django.db import models
from guardian.models import GroupObjectPermissionAbstract, UserObjectPermissionAbstract


class UserObjectPermission(UserObjectPermissionAbstract):
    id = models.BigAutoField(editable=False, unique=True, primary_key=True)
    editable = models.BooleanField(default=True)
    system = models.BooleanField(default=False)

    class Meta(UserObjectPermissionAbstract.Meta):
        abstract = False
        default_related_name = "userobjectpermissions"
        indexes = [
            *UserObjectPermissionAbstract.Meta.indexes,
            models.Index(fields=["content_type", "object_pk", "user"]),
        ]


class GroupObjectPermission(GroupObjectPermissionAbstract):
    id = models.BigAutoField(editable=False, unique=True, primary_key=True)
    editable = models.BooleanField(default=True)

    class Meta(GroupObjectPermissionAbstract.Meta):
        abstract = False
        default_related_name = "groupobjectpermissions"
        indexes = [
            *GroupObjectPermissionAbstract.Meta.indexes,
            models.Index(fields=["content_type", "object_pk", "group"]),
        ]



from enum import Enum


class Button(Enum):
    # Buttons
    REFRESH = "refresh"
    NEW = "new"
    DELETE = "delete"

    # Buttons and Create Buttons
    SAVE = "save"
    SAVE_AND_CLOSE = "save_and_close"
    SAVE_AND_NEW = "save_and_new"

    # Create Buttons
    RESET = "reset"

    # Custom Buttons
    DROPDOWN = "dropdown"
    HYPERLINK = "hyperlink"
    WIDGET = "widget"
    ACTION = "action"

    @classmethod
    def buttons(cls):
        return [
            cls.REFRESH.value,
            cls.NEW.value,
            cls.DELETE.value,
            cls.SAVE.value,
            cls.SAVE_AND_CLOSE.value,
            cls.SAVE_AND_NEW.value,
        ]

    @classmethod
    def create_buttons(cls):
        return [
            cls.SAVE.value,
            cls.SAVE_AND_CLOSE.value,
            cls.SAVE_AND_NEW.value,
            cls.RESET.value,
        ]

    @classmethod
    def custom_buttons(cls):
        return [
            cls.DROPDOWN.value,
            cls.HYPERLINK.value,
            cls.WIDGET.value,
            cls.ACTION.value,
        ]


class ButtonDefaultColor(Enum):
    SUCCESS = "success"
    ERROR = "error"
    WARNING = "warning"
    INFO = "info"
    PRIMARY = "primary"
    SECONDARY = "secondary"
    INHERIT = "inherit"  # 'inherit' means that the button will take the color of a surrounding button group


class ButtonType(Enum):
    DROPDOWN = "dropdown"
    HYPERLINK = "hyperlink"
    WIDGET = "widget"
    ACTION = "action"


class HyperlinkTarget(Enum):
    BLANK = "_blank"
    SELF = "_self"
    PARENT = "_parent"
    TOP = "_top"



from rest_framework.reverse import reverse

from wbcore.contrib.directory.models import BankingContact
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class EmailContactEntryEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcore:directory:entry-emailcontact-list", args=[self.view.kwargs["entry_id"]], request=self.request
        )


class AddressContactEntryEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcore:directory:entry-addresscontact-list", args=[self.view.kwargs["entry_id"]], request=self.request
        )


class SocialMediaContactEntryEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcore:directory:entry-socialmediacontact-list", args=[self.view.kwargs["entry_id"]], request=self.request
        )


class TelephoneContactEntryEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcore:directory:entry-telephonecontact-list", args=[self.view.kwargs["entry_id"]], request=self.request
        )


class WebsiteContactEntryEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcore:directory:entry-websitecontact-list", args=[self.view.kwargs["entry_id"]], request=self.request
        )


class BankingContactEndpointConfig(EndpointViewConfig):
    def get_delete_endpoint(self, **kwargs):
        if self.instance:
            _object = self.view.get_object()
            if _object.status != BankingContact.Status.DRAFT or _object.primary is True:
                return None
        return super().get_delete_endpoint(**kwargs)


class BankingContactEntryEndpointConfig(BankingContactEndpointConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcore:directory:entry-bankingcontact-list", args=[self.view.kwargs["entry_id"]], request=self.request
        )



from .person import (
    SportPersonDisplayConfig,
    PlayerDisplayConfig,
    PlayerTeamDisplayConfig,
    TreeViewPlayerDisplay,
    SportPersonToolTipDisplayConfig,
)
from .league import LeagueDisplayConfig, LeagueSportDisplayConfig
from .team import TeamDisplayConfig, TeamStadiumDisplayConfig
from .stadium import StadiumDisplayConfig
from .sport import SportDisplayConfig
from .season import SeasonDisplayConfig
from .role import RoleDisplayConfig
from .match import MatchDisplayConfig, MatchStadiumDisplayConfig, MatchLeagueDisplayConfig
from .event import (
    EventDisplayConfig,
    EventTypeDisplayConfig,
    EventTypeSportDisplayConfig,
    EventMatchDisplayConfig,
    LeaguePlayerStatisticsDisplayConfig,
    LeagueTeamStatisticsDisplayConfig,
)
from .teamresult import TeamResultsDisplayConfig, TeamResultsLeagueDisplayConfig



import pytest
from dynamic_preferences.registries import global_preferences_registry
from faker import Faker
from rest_framework.test import APIRequestFactory

from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.authentication.models import Permission, User
from wbcore.contrib.directory.models import Entry

from ..factories import (
    BankFactory,
    ClientManagerRelationshipFactory,
    EmployerEmployeeRelationshipFactory,
    PersonFactory,
)
from ..viewsets import (
    CompanyModelViewSet,
    CompanyRepresentationViewSet,
    EntryModelViewSet,
    EntryRepresentationViewSet,
    PersonModelViewSet,
    PersonRepresentationViewSet,
)

fake = Faker()


@pytest.mark.django_db
class TestEntryPermissionQueryset:
    @pytest.fixture
    def entry_fixtures(self):
        """
        Construct the different entries and theirs relationships to the base user
        """
        # True, we create a superuser
        user = UserFactory.create()

        any_person = PersonFactory.create()
        any_company = BankFactory.create()

        employer = EmployerEmployeeRelationshipFactory.create(employee=user.profile).employer
        colleague = EmployerEmployeeRelationshipFactory.create(employer=employer).employee
        client = ClientManagerRelationshipFactory.create(relationship_manager=user.profile).client

        return (
            user,
            User.objects.get(email="AnonymousUser").profile,
            any_person,
            any_company,
            employer,
            colleague,
            client,
        )

    def test_filter_for_internal_user(self, entry_fixtures):
        """
        Test that internal user can see everything
        """
        user, anonymous_profile, any_person, any_company, employer, colleague, client = entry_fixtures
        user.user_permissions.add(
            Permission.objects.get(content_type__app_label="authentication", codename="is_internal_user")
        )
        user = User.objects.get(id=user.id)

        assert set(Entry.objects.filter_for_user(user)) == {
            user.profile.entry_ptr,
            anonymous_profile.entry_ptr,
            any_person.entry_ptr,
            any_company.entry_ptr,
            employer.entry_ptr,
            colleague.entry_ptr,
            client,
        }

    def test_filter_for_normal_user(self, entry_fixtures):
        """
        Test that external/normal user can see:
        - Its own profile
        - Its employer
        - Their colleagues
        - Their clients
        """
        user, anonymous_profile, any_person, any_company, employer, colleague, client = entry_fixtures
        global_preferences_registry.manager()["directory__main_company"] = employer.id

        assert set(Entry.objects.filter_for_user(user)) == {
            user.profile.entry_ptr,
            employer.entry_ptr,
            colleague.entry_ptr,
            client,
        }

    @pytest.mark.parametrize(
        "viewset_class",
        [
            EntryModelViewSet,
            EntryRepresentationViewSet,
            CompanyModelViewSet,
            CompanyRepresentationViewSet,
            PersonModelViewSet,
            PersonRepresentationViewSet,
        ],
    )
    def test_ensure_permission_for_user_on_entry_viewsets(self, entry_fixtures, viewset_class):
        user, anonymous_profile, any_person, any_company, employer, colleague, client = entry_fixtures

        request = APIRequestFactory().get("")
        request.user = user
        viewset = viewset_class(request=request)
        viewset.kwargs = dict()

        # assert for any non employee user
        assert set(viewset.get_queryset()) == set(viewset.model.objects.filter_for_user(user))

        # test for external employee
        global_preferences_registry.manager()["directory__main_company"] = employer.id
        assert set(viewset.get_queryset()) == set(viewset.model.objects.filter_for_user(user))

        # test for internal user
        user.user_permissions.add(
            Permission.objects.get(content_type__app_label="authentication", codename="is_internal_user")
        )
        assert set(viewset.get_queryset()) == set(viewset.model.objects.filter_for_user(user))



import pytest
from rest_framework import serializers
from rest_framework.test import APIRequestFactory

from wbcore.serializers.fields import CharField


@pytest.fixture
def char_field_base():
    field_name = "test_field"
    request = APIRequestFactory().get("")
    default_placeholder = "Enter text here"
    secure_flag = True
    return {
        "field_name": field_name,
        "request": request,
        "default_placeholder": default_placeholder,
        "secure_flag": secure_flag,
    }


class TestCharFieldInitialization:
    """Tests for the initialization of the CharField class."""

    def test_default_initialization(self):
        """Test CharField default initialization with no custom parameters."""
        field = CharField()
        assert field.allow_null is True
        assert field.allow_blank is False
        assert field.placeholder is None
        assert field.secure is False

    def test_custom_initialization(self, char_field_base):
        """Test CharField initialization with custom placeholder and secure parameters."""
        field = CharField(placeholder=char_field_base["default_placeholder"], secure=char_field_base["secure_flag"])
        assert field.placeholder == char_field_base["default_placeholder"]
        assert field.secure is True

    def test_empty_string_placeholder(self, char_field_base):
        """Test that an empty string as a placeholder does not appear in the representation."""
        field = CharField(placeholder="")
        assert field.placeholder == ""
        _, representation = field.get_representation(char_field_base["request"], char_field_base["field_name"])
        assert "placeholder" not in representation

    def test_non_boolean_secure(self, char_field_base):
        """
        Test non-boolean values for the 'secure' attribute.

        Non-boolean values are expected to be treated as truthy (e.g., "no" should behave like True).
        """
        field = CharField(secure="no")
        _, representation = field.get_representation(char_field_base["request"], char_field_base["field_name"])
        assert "secure" in representation
        assert representation["secure"] is True


class TestCharFieldRepresentation:
    """Tests for the representation of the CharField class."""

    def test_representation_without_placeholder_and_secure(self, char_field_base):
        """Test CharField representation without custom placeholder or secure attributes."""
        field = CharField()
        _, representation = field.get_representation(char_field_base["request"], char_field_base["field_name"])
        assert "placeholder" not in representation
        assert "secure" not in representation

    def test_representation_with_placeholder(self, char_field_base):
        """Test CharField representation with a custom placeholder."""
        field = CharField(placeholder="Enter text here")
        _, representation = field.get_representation(char_field_base["request"], char_field_base["field_name"])
        assert "placeholder" in representation
        assert representation["placeholder"] == "Enter text here"

    def test_representation_with_secure(self, char_field_base):
        """Test CharField representation with the secure attribute enabled."""
        field = CharField(secure=True)
        _, representation = field.get_representation(char_field_base["request"], char_field_base["field_name"])
        assert "secure" in representation
        assert representation["secure"] is True

    def test_representation_with_placeholder_and_secure(self, char_field_base):
        """Test CharField representation with both placeholder and secure attributes."""
        field = CharField(placeholder="Enter text here", secure=True)
        _, representation = field.get_representation(char_field_base["request"], char_field_base["field_name"])
        assert "placeholder" in representation
        assert representation["placeholder"] == "Enter text here"
        assert "secure" in representation
        assert representation["secure"] is True


class TestCharFieldValidation:
    """Tests for the validation of empty values in the CharField class."""

    def test_validate_empty_with_allow_null_true(self):
        """Test validation when allow_null is True and data is None."""
        field = CharField(allow_null=True, allow_blank=False)
        should_validate, data = field.validate_empty_values(None)
        assert data == ""
        assert should_validate is False

    def test_validate_empty_with_allow_blank_true(self):
        """Test validation when allow_blank is True and data is None."""
        field = CharField(allow_blank=True, allow_null=False)
        should_validate, data = field.validate_empty_values(None)
        assert data == ""
        assert should_validate is False

    def test_validate_empty_with_allow_null_and_allow_blank_false(self):
        """Test validation when both allow_null and allow_blank are False and data is None."""
        field = CharField(allow_null=False, allow_blank=False)
        with pytest.raises(serializers.ValidationError):
            field.validate_empty_values(None)

    def test_validate_empty_with_allow_null_and_allow_blank_true(self):
        """Test validation when both allow_null and allow_blank are True and data is None."""
        field = CharField(allow_null=True, allow_blank=True)
        should_validate, data = field.validate_empty_values(None)
        assert data == ""
        assert should_validate is False

    def test_validate_empty_with_non_none_data(self):
        """Test validation when data is a non-empty string."""
        field = CharField()
        should_validate, data = field.validate_empty_values("Test")
        assert data == "Test"
        assert should_validate is False

    def test_validate_empty_with_empty_string_data(self):
        """Test validation when data is an empty string."""
        field = CharField(allow_blank=True, allow_null=False)
        should_validate, data = field.validate_empty_values("")
        assert data == ""
        assert should_validate is False

    def test_validate_empty_with_whitespace_string_data(self):
        """Test validation when data is a whitespace string."""
        field = CharField(allow_blank=True, allow_null=False)
        should_validate, data = field.validate_empty_values("   ")
        assert data == "   "
        assert should_validate is False



from wbcore import serializers
from wbcore.contrib.i18n.serializers.fields import TranslationJSONField


class ModelTranslateSerializerMixin(serializers.ModelSerializer):
    _i18n = TranslationJSONField(source="i18n")

    def update(self, instance, validated_data):
        i18n_data = validated_data.pop("i18n", {}) or {}
        if instance.i18n is None:
            instance.i18n = {}
        instance.i18n.update(i18n_data)
        return super().update(instance, validated_data)



import pytest
from wbcore.contrib.color.factories import ColorGradientFactory
from wbcore.contrib.color.models import ColorGradient

COLORS = ["#FFFFFF", "#000000", "#FF5733", "#3357FF", "#FF33A1", "#800080"]


class TestColorGradientModel:
    @pytest.mark.django_db
    def integration_test(self, mocker):
        gradient = ColorGradientFactory()
        assert ColorGradient.objects.filter(id=gradient.id).exists()

    def test_str_method(self, mocker):
        gradient = ColorGradient()
        mocker.patch.object(gradient, "title", "My Gradient")
        assert str(gradient) == "My Gradient"

    def test_default_colors(self):
        gradient = ColorGradient()
        assert gradient.colors == []

    def test_get_gradient(self, mocker):
        gradient = ColorGradient()
        assert list(gradient.get_gradient()) == []

    def test_get_gradient_no_color_argument(self, mocker):
        gradient = ColorGradient()
        mocker.patch.object(gradient, "colors", COLORS)
        palette = list(gradient.get_gradient(color=None, color_size=3))
        assert len(palette) == 3

    def test_get_gradient_with_non_existent_color(self, mocker):
        gradient = ColorGradient()
        mocker.patch.object(gradient, "colors", COLORS)
        non_existent_color = "#123456"
        palette = list(gradient.get_gradient(color=non_existent_color))
        assert len(palette) == len(COLORS)
        assert palette == ["#000000", "#FF5733", "#3357FF", "#FF33A1", "#800080", "#FFFFFF"]

    @pytest.mark.parametrize("color_size", [20, 1, 0, -1])
    def test_get_gradient_with_different_size(self, color_size, mocker):
        gradient = ColorGradient()
        mocker.patch.object(gradient, "colors", COLORS)
        palette = list(gradient.get_gradient(color_size=color_size))
        base_colors = gradient.colors
        assert len(palette) == color_size if color_size > 0 else len(base_colors)
        for index, color in enumerate(palette):
            assert color == base_colors[index % len(base_colors)]

    @pytest.mark.parametrize("color", COLORS)
    def test_get_gradient_with_different_color(self, color, mocker):
        gradient = ColorGradient()
        mocker.patch.object(gradient, "colors", COLORS)
        base_colors = gradient.colors

        color_index = base_colors.index(color)
        palette = list(gradient.get_gradient(color=color))

        assert len(palette) == len(base_colors)
        assert palette == base_colors[color_index:] + base_colors[:color_index]



from typing import TYPE_CHECKING, Iterator

if TYPE_CHECKING:
    from django.db.backends.utils import CursorWrapper


def get_columns(cursor: "CursorWrapper") -> list[str]:
    return [col[0] for col in cursor.description or []]


def dictfetchall[T](cursor: "CursorWrapper", dict_type: type[T] = dict) -> Iterator[T]:
    columns = get_columns(cursor)
    for row in cursor.fetchall():
        # The spec for TypedDict is not compliant with the sepc for dict
        yield dict_type(zip(columns, row, strict=False))  # type: ignore


def dictfetchone(cursor: "CursorWrapper") -> dict:
    columns = get_columns(cursor)
    return dict(zip(columns, cursor.fetchone() or [], strict=False))



from django.urls import include, path
from dynamic_preferences.api.viewsets import GlobalPreferencesViewSet

from wbcore.cache.views import clear_cache
from wbcore.contrib.dynamic_preferences.viewsets import UserPreferencesViewSet
from wbcore.shares.views import ShareAPIView

from .configs.views import ConfigAPIView
from .content_type.viewsets import (
    ContentTypeRepresentationViewSet,
    DynamicObjectIDRepresentationViewSet,
)
from .crontab.viewsets import CrontabScheduleRepresentationViewSet
from .frontend_user_configuration import FrontendUserConfigurationModelViewSet
from .markdown.views import (
    AssetCreateView,
    AssetRetrieveView,
    BlockDiag,
    TemplateTagView,
)
from .menus.views import MenuAPIView
from .metadata.configs.display.views import PresetAPIView
from .release_notes.viewsets import ReleaseNoteReadOnlyModelViewSet
from .reversion.viewsets.viewsets import (
    RevisionModelViewSet,
    RevisionRepresentationViewSet,
    VersionModelViewSet,
    VersionRepresentationViewSet,
    VersionRevisionModelViewSet,
)
from .routers import WBCoreRouter
from .views import Profile

router = WBCoreRouter()
router.register(
    r"frontenduserconfiguration",
    FrontendUserConfigurationModelViewSet,
    basename="frontenduserconfiguration",
)
router.register(r"global_preferences", GlobalPreferencesViewSet, basename="global_preferences")
router.register(r"user_preferences", UserPreferencesViewSet, basename="user_preferences")

router.register(r"version", VersionModelViewSet, basename="version")
router.register(r"revision", RevisionModelViewSet, basename="revision")
router.register(r"revisionrepresentation", RevisionRepresentationViewSet, basename="revisionrepresentation")
router.register(r"versionrepresentation", VersionRepresentationViewSet, basename="versionrepresentation")
router.register(r"releasenote", ReleaseNoteReadOnlyModelViewSet, basename="releasenote")
router.register(r"contenttyperepresentation", ContentTypeRepresentationViewSet, basename="contenttyperepresentation")
router.register(
    r"dynamiccontenttyperepresentation",
    DynamicObjectIDRepresentationViewSet,
    basename="dynamiccontenttyperepresentation",
)
router.register(
    r"crontabschedulerepresentation", CrontabScheduleRepresentationViewSet, basename="crontabschedulerepresentation"
)

revision_router = WBCoreRouter()
revision_router.register(r"version", VersionRevisionModelViewSet, basename="revision-version")


urlpatterns = [
    path("config/", ConfigAPIView.as_view(), name="config"),
    path("profile/", Profile.as_view(), name="profile"),
    path("menu/", MenuAPIView.as_view(), name="menu"),
    path("preset/<str:display_identifier_path>/", PresetAPIView.as_view(), name="preset"),
    path("share/", ShareAPIView.as_view(), name="share"),
    path("markdown/blockdiag/", BlockDiag.as_view(), name="blockdiag"),
    path("markdown/asset/", AssetCreateView.as_view(), name="markdown-asset-upload"),
    path("markdown/asset/<uuid>/", AssetRetrieveView.as_view(), name="asset-retrieve"),
    path("markdown/templatetag/", TemplateTagView.as_view(), name="markdown-tags"),
    path("", include(router.urls)),
    path("revision/<int:revision_id>/", include(revision_router.urls)),
    path(
        "authentication/",
        include(("wbcore.contrib.authentication.urls", "wbcore.contrib.authentication"), namespace="authentication"),
    ),
    path(
        "notifications/",
        include(("wbcore.contrib.notifications.urls", "wbcore.contrib.notifications"), namespace="notifications"),
    ),
    path(
        "io/",
        include(("wbcore.contrib.io.urls", "wbcore.contrib.io"), namespace="io"),
    ),
    path(
        "geography/",
        include(("wbcore.contrib.geography.urls", "wbcore.contrib.geography"), namespace="geography"),
    ),
    path(
        "currency/",
        include(("wbcore.contrib.currency.urls", "wbcore.contrib.currency"), namespace="currency"),
    ),
    path(
        "tags/",
        include(("wbcore.contrib.tags.urls", "wbcore.contrib.tags"), namespace="tags"),
    ),
    path(
        "documents/",
        include(("wbcore.contrib.documents.urls", "wbcore.contrib.documents"), namespace="documents"),
    ),
    path(
        "directory/",
        include(("wbcore.contrib.directory.urls", "wbcore.contrib.directory"), namespace="directory"),
    ),
    path(
        "agenda/",
        include(("wbcore.contrib.agenda.urls", "wbcore.contrib.agenda"), namespace="agenda"),
    ),
    path(
        "workflow/",
        include(("wbcore.contrib.workflow.urls", "wbcore.contrib.workflow"), namespace="workflow"),
    ),
    path(
        "guardian/",
        include(("wbcore.contrib.guardian.urls", "wbcore.contrib.guardian"), namespace="guardian"),
    ),
    path("clear_cache/<str:cache_key>/", clear_cache, name="clear_cache"),
]



from typing import Optional

from django.utils.translation import gettext as _

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig
from wbcore.metadata.configs.display.instance_display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default


class TransitionDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=(
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="from_step", label=_("From")),
                dp.Field(key="to_step", label=_("To")),
                dp.Field(key="icon", label=_("Icon")),
            ),
        )

    def get_instance_display(self) -> Display:
        conditions_section = Section(
            key="conditions_section",
            collapsible=True,
            title=_("Conditions"),
            display=Display(
                pages=[
                    Page(
                        title=_("Conditions"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["conditions_inline"]],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                                inlines=[Inline(key="conditions_inline", endpoint="conditions")],
                            )
                        },
                    ),
                ]
            ),
        )

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["name", "icon"],
                                ["from_step", "to_step"],
                                [
                                    "conditions_section",
                                    "conditions_section",
                                ]
                                if "pk" in self.view.kwargs
                                else [".", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            sections=[conditions_section],
                        ),
                    },
                ),
            ]
        )



from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse

from wbcore import serializers as wb_serializers
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.i18n.serializers.mixins import ModelTranslateSerializerMixin

from ..models import Company, EmployerEmployeeRelationship, Person, Position
from .entries import (
    CompanyRepresentationSerializer,
    EntryModelSerializer,
    SpecializationRepresentationSerializer,
)
from .relationships import PositionRepresentationSerializer


class PersonModelSerializer(ModelTranslateSerializerMixin, EntryModelSerializer):
    primary_employer_repr = wb_serializers.CharField(read_only=True, required=False, label=_("Primary Employer"))
    has_user_account = wb_serializers.CharField(read_only=True, label=_("User Account"))
    name = wb_serializers.CharField(read_only=True)
    last_connection = wb_serializers.DateTimeField(read_only=True, default=None, label=_("Last Connection"))
    personality_profile_red = wb_serializers.RangeSelectField(
        color=WBColor.RED_LIGHT.value, required=False, label=_("Personality Profile Red")
    )
    personality_profile_green = wb_serializers.RangeSelectField(
        color=WBColor.GREEN_LIGHT.value, required=False, label=_("Personality Profile Green")
    )
    personality_profile_blue = wb_serializers.RangeSelectField(
        color=WBColor.BLUE_LIGHT.value, required=False, label=_("Personality Profile Blue")
    )
    _specializations = SpecializationRepresentationSerializer(source="specializations", many=True)
    tier = wb_serializers.CharField(help_text=_("Tier of the primary employer"), label=_("Tier"), required=False)

    def get_user_account_email(self, obj):
        if hasattr(obj, "user_account"):
            return obj.user_account.email

    def get_user_account_last_login(self, obj):
        if hasattr(obj, "user_account"):
            return obj.user_account.last_login

    @wb_serializers.register_resource()
    def employers_list(self, instance, request, user):
        return {"employers": reverse("wbcore:directory:employee-employer-list", args=[instance.id], request=request)}

    @wb_serializers.register_resource()
    def get_clients(self, instance, request, user):
        return {
            "client": f"{reverse('wbcore:directory:clientmanagerrelationship-list', args=[], request=request)}?relationship_manager={instance.id}&status=APPROVED"
        }

    @wb_serializers.register_resource()
    def activity_report(self, instance, request, user):
        return {"activity_report": None}

    def create(self, validated_data):
        instance = super().create(validated_data)

        # When setting a person's employer via the DefaultFromGet in the serializer no save method gets called and the primary mixin doesn't trigger.
        # We need to set the employer as primary manually.
        if instance.employers.exists():
            EmployerEmployeeRelationship.objects.filter(employee=instance, employer=instance.employers.first()).update(
                primary=True
            )
        return instance

    class Meta:
        model = Person
        read_only_fields = (
            "activity_heat",
            "entry_type",
            "primary_employer",
            "initials",
        )
        fields = (
            "id",
            "first_name",
            "last_name",
            "name",
            "computed_str",
            "active_employee",
            "activity_heat",
            "addresses",
            "cities",
            "_cities",
            "birthday",
            "customer_status",
            "entry_type",
            "formal",
            "has_user_account",
            "is_draft_entry",
            "last_connection",
            "personality_profile_blue",
            "personality_profile_green",
            "personality_profile_red",
            "position_in_company",
            "prefix",
            "primary_address",
            "primary_email",
            "primary_manager_repr",
            "primary_employer_repr",
            "primary_telephone",
            "primary_website",
            "primary_social",
            "profile_image",
            "salutation",
            "specializations",
            "_specializations",
            "tier",
            "_additional_resources",
            "initials",
            "description",
            "_i18n",
        )


class NewPersonModelSerializer(PersonModelSerializer):
    primary_employer = wb_serializers.PrimaryKeyRelatedField(
        read_only=False,
        queryset=Company.objects.all(),
        default=wb_serializers.DefaultFromGET("employers"),
        many=False,
        label=_("Primary Employer"),
        required=False,
        allow_null=True,
    )
    _primary_employer = CompanyRepresentationSerializer(
        source="primary_employer",
        many=False,
    )
    position_in_company = wb_serializers.PrimaryKeyRelatedField(
        read_only=False,
        queryset=Position.objects.all(),
        many=False,
        label=_("Position In Company"),
        required=False,
        allow_null=True,
        depends_on=[{"field": "primary_employer", "options": {}}],
    )
    _position_in_company = PositionRepresentationSerializer(
        source="position_in_company",
        many=False,
        depends_on=[{"field": "primary_employer", "options": {}}],
    )
    position_name = wb_serializers.CharField(
        label=_("Position Name"),
        allow_null=True,
        required=False,
        depends_on=[{"field": "primary_employer", "options": {}}],
    )

    def create(self, validated_data: dict) -> Person:
        primary_employer = validated_data.pop("primary_employer", None)
        position_in_company = validated_data.pop("position_in_company", None)
        position_name = validated_data.pop("position_name", None)
        validated_data.pop("employers", None)
        instance = super().create(validated_data)
        if primary_employer:
            EmployerEmployeeRelationship.objects.create(
                employee=instance,
                employer=primary_employer,
                primary=True,
                position=position_in_company,
                position_name=position_name,
            )
        return instance

    class Meta(PersonModelSerializer.Meta):
        model = Person
        read_only_fields = (
            "activity_heat",
            "entry_type",
            "initials",
        )
        fields = PersonModelSerializer.Meta.fields + (
            "_position_in_company",
            "primary_manager",
            "_primary_manager",
            "primary_employer",
            "_primary_employer",
            "position_name",
        )


class PersonModelListSerializer(PersonModelSerializer):
    @wb_serializers.register_resource()
    def get_managers(self, instance, request, user):
        return {
            "manager": f"{reverse('wbcore:directory:clientmanagerrelationship-list', args=[], request=request)}?client={instance.id}"
        }

    class Meta:
        model = Person

        fields = (
            "id",
            "name",
            "activity_heat",
            "addresses",
            "cities",
            "_cities",
            "customer_status",
            "primary_employer_repr",
            "primary_address",
            "primary_email",
            "primary_manager_repr",
            "primary_website",
            "primary_social",
            "primary_telephone",
            "last_event",
            "position_in_company",
            "tier",
            "_additional_resources",
        )



from importlib import import_module
from typing import Callable, Type, Union


def import_from_dotted_path(path: str) -> Union[Type, Callable]:
    *module_path, name = path.split(".")
    return getattr(import_module(".".join(module_path)), name)


def parse_signal_received_for_module(receiver_response) -> tuple[str, any]:
    for receiver, response in receiver_response:
        if response:
            yield receiver.__module__.split(".", 1)[0], response



import re

from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _
from rest_framework import serializers
from rest_framework.reverse import reverse

from .mixins import WBCoreSerializerFieldMixin
from .types import ReturnContentType, WBCoreType


class CharField(WBCoreSerializerFieldMixin, serializers.CharField):
    field_type = WBCoreType.TEXT.value

    def __init__(self, *args, **kwargs):
        self.secure = kwargs.pop("secure", False)
        self.placeholder = kwargs.pop("placeholder", None)
        super().__init__(*args, allow_null=kwargs.pop("allow_null", True), **kwargs)

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)

        if self.placeholder:
            representation["placeholder"] = self.placeholder

        # Add secure attribute if required
        if self.secure:
            representation["secure"] = True

        return key, representation

    def validate_empty_values(self, data):
        if (self.allow_null or self.allow_blank) and data is None:
            data = ""
        return super().validate_empty_values(data)


class TelephoneField(CharField):
    field_type = WBCoreType.TELEPHONE.value


class StringRelatedField(WBCoreSerializerFieldMixin, serializers.StringRelatedField):
    field_type = WBCoreType.TEXT.value


class TextAreaField(CharField):
    field_type = WBCoreType.TEXTAREA.value


class CodeField(CharField):
    field_type = WBCoreType.TEXTAREA.value

    def to_internal_value(self, data):
        try:
            compile(data, "", "exec")
        except Exception as e:
            raise ValidationError(_("Compiling script failed with the exception: {}".format(e))) from e
        return super().to_internal_value(data)


class TextField(CharField):
    field_type = WBCoreType.TEXTEDITOR.value
    texteditor_content_type = ReturnContentType.HTML.value

    def __init__(self, *args, **kwargs):
        """The constructor method.
        This pops the 'plugin_configs' dict from the kwargs and calls the
        parents init method.

        Parameters
        ----------
        plugin_configs : dict, optional
            A dictionary that contains a dictionary for each plugin of the
            texteditor. The configuration for each plugin must be available
            under a unique key.
        """
        self.plugin_configs = kwargs.pop("plugin_configs", None)
        self.disclaimer_length = kwargs.pop("disclaimer_length", None)
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["content_type"] = self.texteditor_content_type
        if not self.plugin_configs:
            representation["plugin_configs"] = {}
        else:
            representation["plugin_configs"] = self.plugin_configs(request)

        if self.disclaimer_length:
            representation["disclaimer_length"] = self.disclaimer_length

        return key, representation


class MarkdownTextField(TextField):
    field_type = WBCoreType.MARKDOWNEDITOR.value
    texteditor_content_type = ReturnContentType.MARKDOWN.value

    def __init__(self, metadata_field=None, *args, **kwargs):
        self.metadata_field = metadata_field
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["image_upload"] = reverse("wbcore:markdown-asset-upload", request=request)
        representation["tags"] = reverse("wbcore:markdown-tags", request=request)
        if self.metadata_field:
            representation["metadata_field"] = self.metadata_field
        return key, representation


class ColorPickerField(CharField):
    field_type = WBCoreType.COLOR.value

    def __init__(self, *args, **kwargs):
        kwargs["copyable"] = True
        super().__init__(*args, **kwargs)

    def to_internal_value(self, data):
        if not re.search(r"^#(?:[0-9a-fA-F]{3}){1,2}$", data):
            raise ValidationError(_("Please provide a valid hexadecimal color."))
        return super().to_internal_value(data)


class URLField(CharField):
    field_type = WBCoreType.URL.value

    def __init__(self, open_internally: bool = False, *args, **kwargs):
        """
        Args:
            open_internally: If true, notify the frontend to open the url as an internal widget
        """
        self.open_internally = open_internally
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["open_internally"] = (
            self.open_internally(self.view) if callable(self.open_internally) else self.open_internally
        )
        return key, representation



from django.conf import settings
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.status import HTTP_500_INTERNAL_SERVER_ERROR
from rest_framework.views import APIView

from .hashes import get_hash_from_user


class GleapUserIdentityAPIView(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request: Request) -> Response:
        return Response(
            (
                str(request.user.id),
                {"email": request.user.email, "name": request.user.get_full_name()},
                get_hash_from_user(request.user.id, settings.GLEAP_IDENTITY_VERIFICATION_SECRET),
            )
        )


class GleapAPITokenAPIView(APIView):
    permission_classes = []

    def get(self, request: Request) -> Response:
        try:
            return Response({"api_token": settings.GLEAP_API_TOKEN})
        except AttributeError:
            return Response({"error": "Gleap is not setup for this server"}, status=HTTP_500_INTERNAL_SERVER_ERROR)



from .e2e_helpers_methods.e2e_helper_methods import (
    click_button_by_label,
    click_close_all_widgets,
    click_element_by_path,
    click_new_button,
    delete_list_entry,
    edit_list_instance,
    fill_out_form_fields,
    find_element,
    find_element_by_text,
    login,
    navigate_to_filter,
    open_create_instance,
    open_menu_item,
    select_async_filter,
    select_column_filter,
    select_filter,
    set_up,
)
from .e2e_helpers_methods.e2e_checks import (
    does_background_color_match,
    is_counter_as_expected,
    is_error_visible,
    is_string_not_visible,
    is_tag_not_visible,
    is_tag_visible,
    is_text_visible,
)

from .tests import GenerateTest, default_config



from django.utils.translation import gettext as _

from wbcore.metadata.configs.titles import TitleViewConfig

from ...models import User


class UserPermissionsModelTitleConfig(TitleViewConfig):
    def get_list_title(self):
        user = User.objects.get(id=self.view.kwargs["user_id"])
        return _("Permissions for {email}").format(email=user.email)


class UserProfileModelTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Your Profile")


class UserModelTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("User: {{email}}")

    def get_list_title(self):
        return _("Users")

    def get_create_title(self):
        return _("New User")



from contextlib import suppress

from django.contrib.auth.models import Group, Permission
from django.db.models import Q, QuerySet
from django.utils.functional import cached_property
from guardian.utils import get_anonymous_user

from wbcore.contrib.authentication.models.users import User


class UserBackend:
    """
    Base UserBackend for basic support (default to all valid user except the anonymous user)
    """

    @cached_property
    def internal_user_permission(self) -> Permission | None:
        try:
            return Permission.objects.get(content_type__app_label="authentication", codename="is_internal_user")
        except Permission.DoesNotExist:
            return None

    def get_internal_groups(self) -> QuerySet[Group]:
        if self.internal_user_permission:
            return Group.objects.filter(permissions=self.internal_user_permission)
        return Group.objects.none()

    def get_internal_users(self) -> QuerySet[User]:
        if self.internal_user_permission:
            qs = User.objects.filter(is_active=True).filter(
                Q(user_permissions=self.internal_user_permission) | Q(groups__in=self.get_internal_groups().all())
            )
            with suppress(User.DoesNotExist):
                qs = qs.exclude(id=get_anonymous_user().id)
            return qs.distinct()
        return User.objects.none()



import pytest
from django.test import Client, TestCase
from django.urls import reverse
from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.example_app.factories import RoleFactory
from wbcore.contrib.example_app.models import Role
from wbcore.contrib.example_app.serializers import RoleModelSerializer
from wbcore.contrib.example_app.tests.test_viewsets.test_utils_viewsets import (
    get_create_view,
    get_delete_view,
    get_detail_view,
    get_partial_view,
    get_update_view,
)
from wbcore.contrib.example_app.viewsets import RoleModelViewSet


@pytest.mark.django_db
class TestRoleModelViewSet(TestCase):
    def setUp(self):
        self.user = SuperUserFactory()
        self.client = Client()
        self.client.force_login(user=self.user)
        self.list_url = reverse("example_app:role-list")
        self.detail_url_str = "example_app:role-detail"

    def test_list_view(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

    def test_create_view(self):
        role = RoleFactory()
        response = get_create_view(self.client, role, self.user, self.list_url, RoleModelViewSet)
        self.assertEqual(response.status_code, 201)
        self.assertTrue(Role.objects.filter(title=role.title).exists())

    def test_detail_view(self):
        instance = RoleFactory()
        response = get_detail_view(self.client, instance.pk, self.detail_url_str)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["title"], instance.title)

    def test_update_view(self):
        instance = RoleFactory()
        instance.title = "Updated Instance"
        response = get_update_view(self.client, instance, RoleModelSerializer, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["title"], instance.title)

    def test_partial_update_view(self):
        instance = RoleFactory()
        response = get_partial_view(self.client, instance.id, {"title": "Patched Title"}, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["title"], instance.title)

    def test_delete_view(self):
        instance = RoleFactory()
        response = get_delete_view(self.client, self.detail_url_str, instance.pk)
        self.assertEqual(response.status_code, 204)
        self.assertFalse(Role.objects.filter(pk=instance.pk).exists())

    def test_ordering_fields(self):
        first_role, second_role, third_role = "Role A", "Role B", "Role C"
        RoleFactory(title=second_role)
        RoleFactory(title=first_role)
        RoleFactory(title=third_role)

        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["count"], Role.objects.count())
        self.assertEqual(response.data["results"][0]["title"], first_role)
        self.assertEqual(response.data["results"][1]["title"], second_role)
        self.assertEqual(response.data["results"][2]["title"], third_role)



from django_celery_beat.models import CrontabSchedule

from wbcore import serializers


class CrontabScheduleRepresentationSerializer(serializers.RepresentationSerializer):
    endpoint = "wbcore:crontabschedulerepresentation-list"
    value_key = "id"

    human_readable = serializers.SerializerMethodField()
    obj_str = serializers.SerializerMethodField()

    def get_obj_str(self, obj):
        return str(obj)

    def get_human_readable(self, obj):
        return obj.human_readable

    def __init__(self, *args, label_key="{{human_readable}} [ {{obj_str}} ]", **kwargs):
        super().__init__(*args, label_key=label_key, **kwargs)

    class Meta:
        model = CrontabSchedule
        fields = ("id", "month_of_year", "day_of_month", "day_of_week", "hour", "minute", "human_readable", "obj_str")



import logging
from datetime import date, timedelta

import pandas as pd
import plotly.graph_objects as go
from django.db.models import (
    DurationField,
    ExpressionWrapper,
    F,
    OuterRef,
    Subquery,
    Sum,
    Value,
)
from django.db.models.fields import FloatField
from django.db.models.functions import Coalesce, Concat, ExtractMinute
from django.utils.translation import gettext as _
from rest_framework import filters

from wbcore import viewsets
from wbcore.filters import DjangoFilterBackend

from .. import serializers
from ..filters import UserActivityChartFilter
from ..models import User, UserActivity
from ..serializers import UserActivityTableSerializer
from .display.user_activities import (
    UserActivityModelDisplay,
    UserActivityModelUserDisplay,
    UserActivityTableDisplayConfig,
)
from .endpoints.user_activities import (
    UserActivityModelEndpointConfig,
    UserActivityTableEndpointConfig,
    UserActivityUserModelEndpointConfig,
)
from .titles.user_activities import (
    UserActivityChartTitleConfig,
    UserActivityModelTitleConfig,
    UserActivityTableTitleConfig,
    UserActivityUserChartTitleConfig,
    UserActivityUserModelTitleConfig,
)

logger = logging.getLogger()


class UserActivityModelViewSet(viewsets.ModelViewSet):
    filterset_fields = {
        "IP": ["exact", "icontains"],
        "user": ["exact"],
        "date": ["gte", "exact", "lte"],
        "latest_refresh": ["gte", "exact", "lte"],
        "user_agent_info": ["exact", "icontains"],
    }

    filter_backends = (filters.SearchFilter, filters.OrderingFilter, DjangoFilterBackend)
    search_fields = ("user__email", "user_agent_info")
    ordering_fields = ("date", "latest_refresh", "time_online_minute")
    ordering = ["-date"]

    queryset = UserActivity.objects.all()
    serializer_class = serializers.UserActivitySerializer

    display_config_class = UserActivityModelDisplay
    endpoint_config_class = UserActivityModelEndpointConfig
    title_config_class = UserActivityModelTitleConfig

    def get_queryset(self):
        expression = ExpressionWrapper(F("latest_refresh") - F("date"), output_field=DurationField())
        return UserActivity.objects.annotate(time_online_minute=ExtractMinute(expression))


class UserActivityUserModelViewSet(UserActivityModelViewSet):
    display_config_class = UserActivityModelUserDisplay
    endpoint_config_class = UserActivityUserModelEndpointConfig
    title_config_class = UserActivityUserModelTitleConfig

    filter_backends = (filters.SearchFilter, DjangoFilterBackend, filters.OrderingFilter)
    filterset_fields = {
        "IP": ["exact"],
        "date": ["gte", "exact", "lte"],
        "latest_refresh": ["gte", "exact", "lte"],
        "user_agent_info": ["exact", "icontains"],
    }

    def get_queryset(self):
        user = User.objects.get(id=self.kwargs["user_id"])
        return super().get_queryset().filter(user=user)


class UserActivityTable(viewsets.ModelViewSet):
    queryset = UserActivity.objects.all()
    filter_backends = (filters.SearchFilter, filters.OrderingFilter)
    serializer_class = UserActivityTableSerializer
    ordering_fields = ["user_repr", "today_activity", "yesterday_activity", "beforeyesterday_activity"]
    ordering = ["-today_activity"]
    search_fields = ["user_repr"]

    endpoint_config_class = UserActivityTableEndpointConfig
    title_config_class = UserActivityTableTitleConfig
    display_config_class = UserActivityTableDisplayConfig

    def get_queryset(self):
        expression = ExpressionWrapper(F("latest_refresh") - F("date"), output_field=DurationField())

        def get_time_day(day):
            return Coalesce(
                Subquery(
                    UserActivity.objects.filter(user__id=OuterRef("pk"), date__date=day)
                    .annotate(_time_online_minute=ExtractMinute(expression))
                    .values("user")
                    .annotate(total_time=Sum("_time_online_minute"))
                    .values("total_time")[:1],
                    output_field=FloatField(),
                ),
                0.0,
            )

        today = date.today()
        qs = User.objects.annotate(
            beforeyesterday_activity=get_time_day(today - timedelta(days=2)),
            today_activity=get_time_day(today),
            yesterday_activity=get_time_day(today - timedelta(days=1)),
            user_repr=F("profile__computed_str"),
        )
        return qs


class UserActivityChart(viewsets.ChartViewSet):
    queryset = UserActivity.objects.all()
    title_config_class = UserActivityChartTitleConfig
    filter_backends = (DjangoFilterBackend,)
    filterset_class = UserActivityChartFilter

    def get_queryset(self):
        return UserActivity.objects.all().annotate(
            username=Concat(F("user__profile__first_name"), Value(" "), F("user__profile__last_name"))
        )

    def get_plotly(self, queryset):
        fig = go.Figure()
        if queryset.count() > 0:
            df = pd.DataFrame(queryset.order_by("-date").values("latest_refresh", "date", "username"))

            df.date = pd.to_datetime(df.date).dt.round("Min")
            df.latest_refresh = pd.to_datetime(df.latest_refresh).dt.round("Min")
            min_date = df.date.min()
            max_date = df.latest_refresh.max()
            freq = "H"
            if (max_date - min_date).total_seconds() < 86400:
                freq = "D"
            min_date = min_date.floor(freq)
            max_date = max_date.ceil(freq)

            for username, dff in df.groupby("username"):
                cum_times = pd.Series(
                    dt
                    for group in [
                        pd.date_range(start, end, freq="Min")
                        for start, end in zip(dff.date, dff.latest_refresh, strict=False)
                    ]
                    for dt in group
                ).value_counts()
                fig.add_trace(
                    go.Histogram(
                        histfunc="sum",
                        xbins=dict(start=min_date, end=max_date, size=3600000),  # bins used for histogram
                        y=cum_times.values,
                        x=cum_times.index,
                        bingroup=1,
                        name=username,
                    ),
                )
            fig.update_layout(
                barmode="stack",
                paper_bgcolor="rgba(0,0,0,0)",
                plot_bgcolor="rgba(0,0,0,0)",
                title=_("<b>User Activity</b>"),
                xaxis=dict(title=_("Time"), type="date", range=[min_date, max_date]),
                yaxis=dict(title=_("Cummulative time"), type="linear"),
                updatemenus=[
                    dict(
                        buttons=[
                            # dict(
                            #     args = ['xbins.size', 60000],
                            #     label = 'Minute',
                            #     method = 'restyle',
                            # ),
                            dict(
                                args=["xbins.size", 3600000],
                                label=_("Hour"),
                                method="restyle",
                            ),
                            dict(
                                args=["xbins.size", 86400000],
                                label=_("Day"),
                                method="restyle",
                            ),
                            dict(
                                args=["xbins.size", 604800000],
                                label=_("Week"),
                                method="restyle",
                            ),
                            dict(
                                args=["xbins.size", "M1"],
                                label=_("Month"),
                                method="restyle",
                            ),
                        ]
                    )
                ],
                autosize=True,
            )

        return fig


class UserActivityUserChart(UserActivityChart):
    title_config_class = UserActivityUserChartTitleConfig

    def get_queryset(self):
        user = User.objects.get(id=self.kwargs["user_id"])
        return super().get_queryset().filter(user=user)



from reportlab.platypus import Paragraph as BaseParagraph


class FormattedParagraph(BaseParagraph):
    def __init__(self, text, *args, **kwargs):
        text = text.replace("<br>", "<br/>")  # convert the HTML line break into a compatible XML line break tag
        super().__init__(text, *args, **kwargs)



from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from wbcore import filters as wb_filters
from wbcore.contrib.documents.models import (
    Document,
    DocumentType,
    ShareableLink,
    ShareableLinkAccess,
)


class ShareableLinkFilter(wb_filters.FilterSet):
    valid = wb_filters.BooleanFilter(label=_("Valid"), method="boolean_is_valid")
    valid_until = wb_filters.DateTimeRangeFilter(
        label="Valid Until",
    )
    link = wb_filters.CharFilter(label=_("Link"), method="filter_uuid")

    def filter_uuid(self, queryset, name, value):
        if value:
            return queryset.filter(Q(uuid__icontains=value))
        return queryset

    def boolean_is_valid(self, queryset, name, value):
        if value is not None:
            return queryset.filter(valid=value)
        return queryset

    class Meta:
        model = ShareableLink
        fields = {"one_time_link": ["exact"], "document": ["exact"]}


class ShareableLinkAccessFilter(wb_filters.FilterSet):
    clearable = (False,)
    accessed = wb_filters.DateTimeRangeFilter(
        label="Accessed between date",
    )

    class Meta:
        model = ShareableLinkAccess
        fields = {"shareable_link": ["exact"]}


class DocumentFilter(wb_filters.FilterSet):
    document_type = wb_filters.ModelChoiceFilter(
        label=_("Document Types"),
        queryset=DocumentType.objects.all(),
        endpoint=DocumentType.get_representation_endpoint(),
        value_key=DocumentType.get_representation_value_key(),
        label_key=DocumentType.get_representation_label_key(),
        method="filter_document_type",
    )

    def filter_document_type(self, queryset, name, value):
        if value:
            return queryset.filter(document_type__in=value.get_descendants(include_self=True))
        return queryset

    class Meta:
        model = Document
        fields = {"name": ["exact", "icontains"], "system_created": ["exact"], "updated": ["lte", "gte"]}


class DocumentTypeFilter(wb_filters.FilterSet):
    document_count__lte = wb_filters.NumberFilter(
        label=_("Document Count"),
        field_name="document_count",
        lookup_expr="lte",
    )

    document_count__gte = wb_filters.NumberFilter(
        label=_("Document Count"),
        field_name="document_count",
        lookup_expr="gte",
    )

    class Meta:
        model = DocumentType
        fields = {"parent": ["exact"], "name": ["exact", "icontains"]}



from django.utils.translation import gettext as _

from wbcore.metadata.configs.titles import TitleViewConfig


class CustomerStatusTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Customer Statuses")

    def get_create_title(self):
        return _("New Customer Status")

    def get_instance_title(self):
        return _("Customer Status")


class PositionTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Company Positions")

    def get_create_title(self):
        return _("New Company Position")

    def get_instance_title(self):
        return _("Company Position")


class CompanyTypeTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Company Types")

    def get_create_title(self):
        return _("New Company Type")

    def get_instance_title(self):
        return _("Company Type")


class SpecializationTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Specializations")

    def get_create_title(self):
        return _("New Specialization")

    def get_instance_title(self):
        return _("Specialization")



# Generated by Django 5.0.6 on 2024-08-06 07:28

import wbcore.contrib.example_app.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("example_app", "0004_alter_player_transfer_value"),
    ]

    operations = [
        migrations.AddField(
            model_name="sportperson",
            name="profile_image",
            field=models.ImageField(
                blank=True,
                max_length=256,
                null=True,
                upload_to=wbcore.contrib.example_app.models.upload_to_profile_images,
            ),
        ),
    ]



from django.db.models import F, Q, QuerySet
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from rest_framework import permissions
from rest_framework.decorators import action
from rest_framework.response import Response

from wbcore import viewsets
from wbcore.contrib.authentication.models import User
from wbcore.contrib.workflow.filters import (
    AssignedProcessStepFilter,
    ProcessFilter,
    ProcessStepFilter,
    ProcessStepProcessFilter,
    ProcessStepStepFilter,
    ProcessWorkflowFilter,
)
from wbcore.contrib.workflow.models import (
    Process,
    ProcessStep,
    Step,
    Transition,
    UserStep,
)
from wbcore.contrib.workflow.serializers import (
    AssignedProcessStepSerializer,
    ProcessModelSerializer,
    ProcessStepModelSerializer,
)
from wbcore.contrib.workflow.viewsets.display import (
    AssignedProcessStepDisplayConfig,
    ProcessDisplayConfig,
    ProcessStepDisplayConfig,
)
from wbcore.contrib.workflow.viewsets.endpoints import (
    AssignedProcessStepEndpointConfig,
    ProcessEndpointConfig,
)
from wbcore.contrib.workflow.viewsets.titles import (
    AssignedProcessStepTitleConfig,
    ProcessStepTitleConfig,
    ProcessTitleConfig,
)


class ProcessModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/process.md"
    queryset = Process.objects.all()
    serializer_class = ProcessModelSerializer
    display_config_class = ProcessDisplayConfig
    title_config_class = ProcessTitleConfig
    endpoint_config_class = ProcessEndpointConfig
    search_fields = ("workflow__name", "content_type__model")
    ordering = ("workflow__name", "state", "content_type__model", "id")
    ordering_fields = (
        "id",
        "workflow__name",
        "started",
        "finished",
        "content_type__model",
    )

    def get_queryset(self):
        if workflow_id := self.kwargs.get("workflow_id"):
            return super().get_queryset().filter(workflow=workflow_id)
        return super().get_queryset()

    def get_filterset_class(self, request):
        if self.kwargs.get("workflow_id"):
            return ProcessWorkflowFilter
        return ProcessFilter


class ProcessStepModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/processstep.md"
    queryset = ProcessStep.objects.all()
    serializer_class = ProcessStepModelSerializer
    title_config_class = ProcessStepTitleConfig
    endpoint_config_class = ProcessEndpointConfig
    display_config_class = ProcessStepDisplayConfig
    serializer_class = ProcessStepModelSerializer
    search_fields = ("process__workflow__name", "step__name")
    ordering = ("step__name", "state", "process__id", "id")
    ordering_fields = (
        "id",
        "process__id",
        "step__name",
        "started",
        "finished",
        "assignee__computed_str",
        "group__name",
        "permission__name",
        "status",
        "error_message",
    )

    @cached_property
    def request_user(self) -> User:
        return self.request.user

    def get_queryset(self) -> QuerySet[ProcessStep]:
        queryset: QuerySet[ProcessStep] = (
            super().get_queryset().select_related("assignee", "permission", "group", "process")
        )
        if not self.request_user.is_superuser:
            queryset = queryset.filter(
                Q(permission__user=self.request_user)
                | Q(permission__group__user=self.request_user)
                | Q(permission__isnull=True)
            )
        if process_id := self.kwargs.get("process_id"):
            queryset = queryset.filter(process=process_id)
        elif step_id := self.kwargs.get("step_id"):
            queryset = queryset.filter(step=step_id)
        return queryset

    def get_filterset_class(self, request):
        if self.kwargs.get("process_id"):
            return ProcessStepProcessFilter
        if self.kwargs.get("step_id"):
            return ProcessStepStepFilter
        return ProcessStepFilter

    @action(detail=True, methods=["PATCH"], permission_classes=[permissions.IsAuthenticated])
    def next(self, request, pk=None):
        process_step = get_object_or_404(ProcessStep, pk=pk)
        transition = get_object_or_404(Transition, pk=request.GET.get("transition_id"))
        Step.start_next_step(process_step, transition)
        return Response({"status": "ok"})


class AssignedProcessStepModelViewSet(ProcessStepModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/assignedprocessstep.md"
    title_config_class = AssignedProcessStepTitleConfig
    serializer_class = AssignedProcessStepSerializer
    endpoint_config_class = AssignedProcessStepEndpointConfig
    display_config_class = AssignedProcessStepDisplayConfig
    search_fields = ("process__workflow__name", "step__name")
    ordering = ("step__name", "step__workflow__name", "process__id", "id")
    ordering_fields = (
        "step__name",
        "started",
        "finished",
        "group__name",
        "permission__name",
        "status",
        "attached_model",
        "workflow_name",
    )

    def get_queryset(self) -> QuerySet[ProcessStep]:
        qs = (
            super()
            .get_queryset()
            .filter(
                Q(step__step_type=Step.StepType.USERSTEP)
                & Q(state=ProcessStep.StepState.ACTIVE)
                & Q(
                    Q(assignee=self.request_user)
                    | Q(
                        Q(group__user=self.request_user)
                        & Q(step__in=UserStep.objects.filter(assignee_method__isnull=True))
                        & Q(assignee__isnull=True)
                    )
                )
            )
            .annotate(
                attached_model=F("step__workflow__model__model"),
                workflow_name=F("step__workflow__name"),
            )
        ).select_related("step__workflow")
        return qs

    def get_filterset_class(self, request):
        return AssignedProcessStepFilter
